"""
[[https://shop-eu.emfit.com/products/emfit-qs][Emfit QS]] sleep tracker

Consumes data exported by https://github.com/karlicoss/emfitexport
"""

REQUIRES = [
    'git+https://github.com/karlicoss/emfitexport',
]

from contextlib import contextmanager
import dataclasses
from datetime import datetime, time, timedelta
import inspect
from pathlib import Path
from typing import Any, Dict, Iterable, Iterator, List, Optional

from my.core import (
    get_files,
    stat,
    Res,
    Stats,
)
from my.core.common import mcachew
from my.core.cachew import cache_dir
from my.core.error import set_error_datetime, extract_error_datetime
from my.core.pandas import DataFrameT

from my.config import emfit as config

import emfitexport.dal as dal


Emfit = dal.Emfit


# TODO move to common?
def dir_hash(path: Path):
    mtimes = tuple(p.stat().st_mtime for p in get_files(path, glob='*.json'))
    return mtimes


def _cachew_depends_on():
    return dir_hash(config.export_path)


# TODO take __file__ into account somehow?
@mcachew(cache_path=cache_dir() / 'emfit.cache', depends_on=_cachew_depends_on)
def datas() -> Iterable[Res[Emfit]]:
    # data from emfit is coming in UTC. There is no way (I think?) to know the 'real' timezone, and local times matter more for sleep analysis
    # TODO actually this is wrong?? there is some sort of local offset in the export
    emfit_tz = config.timezone

    ## backwards compatibility (old DAL didn't have cpu_pool argument)
    cpu_pool_arg = 'cpu_pool'
    pass_cpu_pool = cpu_pool_arg in inspect.signature(dal.sleeps).parameters
    if pass_cpu_pool:
        from my.core._cpu_pool import get_cpu_pool

        kwargs = {cpu_pool_arg: get_cpu_pool()}
    else:
        kwargs = {}
    ##

    for x in dal.sleeps(config.export_path, **kwargs):
        if isinstance(x, Exception):
            yield x
        else:
            if x.sid in config.excluded_sids:
                # TODO should be responsibility of export_path (similar to HPI?)
                continue
            # TODO maybe have a helper to 'patch up' all dattetimes in a namedtuple/dataclass?
            # TODO do the same for jawbone data?
            # fmt: off
            x = dataclasses.replace(
                x,
                start      =x.start      .astimezone(emfit_tz),
                end        =x.end        .astimezone(emfit_tz),
                sleep_start=x.sleep_start.astimezone(emfit_tz),
                sleep_end  =x.sleep_end  .astimezone(emfit_tz),
            )
            # fmt: on
            yield x


# TODO should be used for jawbone data as well?
def pre_dataframe() -> Iterable[Res[Emfit]]:
    # TODO shit. I need some sort of interrupted sleep detection?
    g: List[Emfit] = []

    def flush() -> Iterable[Res[Emfit]]:
        if len(g) == 0:
            return
        elif len(g) == 1:
            r = g[0]
            g.clear()
            yield r
        else:
            err = RuntimeError(f'Multiple sleeps per night, not supported yet: {g}')
            set_error_datetime(err, dt=datetime.combine(g[0].date, time.min))
            g.clear()
            yield err

    for x in datas():
        if isinstance(x, Exception):
            yield x
            continue
        # otherwise, Emfit
        if len(g) != 0 and g[-1].date != x.date:
            yield from flush()
        g.append(x)
    yield from flush()


def dataframe() -> DataFrameT:
    dicts: List[Dict[str, Any]] = []
    last: Optional[Emfit] = None
    for s in pre_dataframe():
        d: Dict[str, Any]
        if isinstance(s, Exception):
            edt = extract_error_datetime(s)
            d = {
                'date': edt,
                'error': str(s),
            }
        else:
            dd = s.date
            pday = dd - timedelta(days=1)
            if last is None or last.date != pday:
                hrv_change = None
            else:
                # todo it's change during the day?? dunno if reasonable metric
                hrv_change = s.hrv_evening - last.hrv_morning
            # todo maybe changes need to be handled in a more generic way?

            # todo ugh. get rid of hardcoding, just generate the schema automatically
            # TODO use 'workdays' provider....
            # fmt: off
            d = {
                'date'       : dd,

                'sleep_start': s.sleep_start,
                'sleep_end'  : s.sleep_end,
                'bed_time'   : s.time_in_bed, # eh, this is derived frop sleep start / end. should we compute it on spot??

                # these are emfit specific
                'coverage'   : s.sleep_hr_coverage,
                'avg_hr'     : s.measured_hr_avg,
                'hrv_evening': s.hrv_evening,
                'hrv_morning': s.hrv_morning,
                'recovery'   : s.recovery,
                'hrv_change' : hrv_change,
                'respiratory_rate_avg': s.respiratory_rate_avg,
            }
            # fmt: on
            last = s  # meh
        dicts.append(d)

    import pandas

    return pandas.DataFrame(dicts)


def stats() -> Stats:
    return stat(pre_dataframe)


@contextmanager
def fake_data(nights: int = 500) -> Iterator:
    from my.core.cfg import tmp_config
    from tempfile import TemporaryDirectory

    import pytz

    with TemporaryDirectory() as td:
        tdir = Path(td)
        gen = dal.FakeData()
        gen.fill(tdir, count=nights)

        class override:
            class emfit:
                export_path = tdir
                excluded_sids = ()
                timezone = pytz.timezone('Europe/London')  # meh

        with tmp_config(modules=__name__, config=override) as cfg:
            yield cfg


# TODO remove/deprecate it? I think used by timeline
def get_datas() -> List[Emfit]:
    # todo ugh. run lint properly
    return list(sorted(datas(), key=lambda e: e.start))  # type: ignore


# TODO move away old entries if there is a diff??



from pathlib import Path

# todo preinit isn't really a good name? it's only in a separate file because
# - it's imported from my.core.init (so we wan't to keep this file as small/reliable as possible, hence not common or something)
# - we still need this function in __main__, so has to be separate from my/core/init.py
def get_mycfg_dir() -> Path:
    import appdirs  # type: ignore[import-untyped]
    import os
    # not sure if that's necessary, i.e. could rely on PYTHONPATH instead
    # on the other hand, by using MY_CONFIG we are guaranteed to load it from the desired path?
    mvar = os.environ.get('MY_CONFIG')
    if mvar is not None:
        mycfg_dir = Path(mvar)
    else:
        mycfg_dir = Path(appdirs.user_config_dir('my'))
    return mycfg_dir



REQUIRES = ["git+https://github.com/seanbreckenridge/ipgeocache"]

from .fallback.via_ip import *

from my.core.warnings import high

high("my.location.via_ip is deprecated, use my.location.fallback.via_ip instead")



from my.tests.common import skip_if_not_karlicoss as pytestmark


def test() -> None:
    from my.emfit import datas
    # TODO this should be implement via sanity checks/stat instead?
    # the same code will be used for tests & for user reports
    ds = [x for x in datas() if not isinstance(x, Exception)]
    for d in ds:
        assert d.start.tzinfo is not None
        assert d.end.tzinfo is not None
        assert d.sleep_start.tzinfo is not None
        assert d.sleep_end.tzinfo is not None


def test_tz() -> None:
    from my.emfit import datas
    # TODO check errors too?
    ds = [x for x in datas() if not isinstance(x, Exception)]

    # this was winter time, so GMT, UTC+0
    [s0109] = [s for s in ds if s.date.strftime('%Y%m%d') == '20190109']
    assert s0109.end.strftime('%H:%M') == '06:42'

    # summer time, so UTC+1
    [s0411] = [s for s in ds if s.date.strftime('%Y%m%d') == '20190411']
    assert s0411.end.strftime('%H:%M') == '09:30'



'''
Timezone data provider, used to localize timezone-unaware timestamps for other modules
'''
from datetime import datetime
from ...core.common import tzdatetime

# todo hmm, kwargs isn't mypy friendly.. but specifying types would require duplicating default args. uhoh
def localize(dt: datetime, **kwargs) -> tzdatetime:
    # todo document patterns for combining multiple data sources
    # e.g. see https://github.com/karlicoss/HPI/issues/89#issuecomment-716495136
    from . import via_location as L
    from .common import localize_with_policy
    return localize_with_policy(L.localize, dt, **kwargs)



"""
[[https://www.goodreads.com][Goodreads]] statistics
"""
REQUIRES = [
    'git+https://github.com/karlicoss/goodrexport',
]


from dataclasses import dataclass
from my.core import Paths
from my.config import goodreads as user_config

@dataclass
class goodreads(user_config):
    # paths[s]/glob to the exported JSON data
    export_path: Paths

from my.core.cfg import make_config, Attrs

def _migration(attrs: Attrs) -> Attrs:
    export_dir = 'export_dir'
    if export_dir in attrs: # legacy name
        attrs['export_path'] = attrs[export_dir]
        from my.core.warnings import high
        high(f'"{export_dir}" is deprecated! Please use "export_path" instead."')
    return attrs
config = make_config(goodreads, migration=_migration)

#############################3


from my.core import get_files
from typing import Sequence, Iterator
from pathlib import Path

def inputs() -> Sequence[Path]:
    return get_files(config.export_path)


from datetime import datetime
import pytz


from goodrexport import dal


def _dal() -> dal.DAL:
    return dal.DAL(inputs())


def reviews() -> Iterator[dal.Review]:
    return _dal().reviews()


# todo should be in DAL?
def books() -> Iterator[dal.Book]:
    for r in reviews():
        yield r.book


#######
# todo ok, not sure these really belong here...

from my.core.common import datetime_aware
@dataclass
class Event:
    dt: datetime_aware
    summary: str
    eid: str


def events() -> Iterator[Event]:
    for b in books():
        yield Event(
            dt=b.date_added,
            summary=f'Added book "{b.title}"', # todo shelf?
            eid=b.id
        )
    # todo finished? other updates?


def print_read_history() -> None:
    def ddate(x):
        if x is None:
            return datetime.fromtimestamp(0, pytz.utc)
        else:
            return x

    def key(b):
        return ddate(b.date_started)

    def fmtdt(dt):
        if dt is None:
            return dt
        tz = pytz.timezone('Europe/London')
        return dt.astimezone(tz)
    for b in sorted(books(), key=key):
        print(f"""
{b.title} by {', '.join(b.authors)}
    started : {fmtdt(b.date_started)}
    finished: {fmtdt(b.date_read)}
        """)



"""
[[https://uk.kobobooks.com/products/kobo-aura-one][Kobo]] e-ink reader: annotations and reading stats
"""
from __future__ import annotations

REQUIRES = [
    'kobuddy',
]

from dataclasses import dataclass
from typing import Iterator

from my.core import (
    get_files,
    stat,
    Paths,
    Stats,
)
from my.core.cfg import make_config
import my.config

import kobuddy
from kobuddy import Highlight, get_highlights
from kobuddy import *


@dataclass
class kobo(my.config.kobo):
    '''
    Uses [[https://github.com/karlicoss/kobuddy#as-a-backup-tool][kobuddy]] outputs.
    '''

    # path[s]/glob to the exported databases
    export_path: Paths


config = make_config(kobo)

# TODO not ideal to set it here.. should switch kobuddy to use a proper DAL
kobuddy.DATABASES = list(get_files(config.export_path))


def highlights() -> Iterator[Highlight]:
    return kobuddy._iter_highlights()


def stats() -> Stats:
    return stat(highlights)


## TODO hmm. not sure if all this really belongs here?... perhaps orger?


from typing import Callable, Union, List

# TODO maybe type over T?
_Predicate = Callable[[str], bool]
Predicatish = Union[str, _Predicate]


def from_predicatish(p: Predicatish) -> _Predicate:
    if isinstance(p, str):

        def ff(s):
            return s == p

        return ff
    else:
        return p


def by_annotation(predicatish: Predicatish, **kwargs) -> List[Highlight]:
    pred = from_predicatish(predicatish)

    res: List[Highlight] = []
    for h in get_highlights(**kwargs):
        if pred(h.annotation):
            res.append(h)
    return res


def get_todos() -> List[Highlight]:
    def with_todo(ann):
        if ann is None:
            ann = ''
        return 'todo' in ann.lower().split()

    return by_annotation(with_todo)


def test_todos() -> None:
    todos = get_todos()
    assert len(todos) > 3



import os
from subprocess import check_call
import sys


def test_lists_modules() -> None:
    # hack PYTHONUTF8 for windows
    # see  https://github.com/karlicoss/promnesia/issues/274
    # https://memex.zulipchat.com/#narrow/stream/279600-promnesia/topic/indexing.3A.20utf8.28emoji.29.20filenames.20in.20Windows
    # necessary for this test cause emooji is causing trouble
    # TODO need to fix it properly
    env = {
        **os.environ,
        'PYTHONUTF8': '1',
    }
    check_call([sys.executable, '-m', 'my.core', 'modules'], env=env)



"""
[[https://github.com/burtonator/polar-bookshelf][Polar]] articles and highlights
"""
from pathlib import Path
from typing import cast, TYPE_CHECKING


import my.config

# todo use something similar to tz.via_location for config fallback
if not TYPE_CHECKING:
    user_config = getattr(my.config, 'polar', None)
else:
    # mypy can't handle dynamic base classes... https://github.com/python/mypy/issues/2477
    user_config = object

# by default, Polar doesn't need any config, so perhaps makes sense to make it defensive here
if user_config is None:
    class user_config: # type: ignore[no-redef]
        pass


from .core import PathIsh
from dataclasses import dataclass
@dataclass
class polar(user_config):
    '''
    Polar config is optional, you only need it if you want to specify custom 'polar_dir'
    '''
    polar_dir: PathIsh = Path('~/.polar').expanduser()
    defensive: bool = True # pass False if you want it to fail faster on errors (useful for debugging)


from .core import make_config
config = make_config(polar)

# todo not sure where it keeps stuff on Windows?
# https://github.com/burtonator/polar-bookshelf/issues/296

from datetime import datetime
from typing import List, Dict, Iterable, NamedTuple, Sequence, Optional
import json

from .core import LazyLogger, Json, Res
from .core.compat import fromisoformat
from .core.error import echain, sort_res_by
from .core.konsume import wrap, Zoomable, Wdict


logger = LazyLogger(__name__)


# Ok I guess handling comment-level errors is a bit too much..
Cid = str
class Comment(NamedTuple):
    cid: Cid
    created: datetime
    text: str

Hid = str
class Highlight(NamedTuple):
    hid: Hid
    created: datetime
    selection: str
    comments: Sequence[Comment]
    tags: Sequence[str]
    page: int  # 1-indexed
    color: Optional[str] = None


Uid = str
class Book(NamedTuple):
    created: datetime
    uid: Uid
    path: Path
    title: Optional[str]
    # TODO hmmm. I think this needs to be defensive as well...
    # think about it later.
    items: Sequence[Highlight]

    tags: Sequence[str]

    @property
    def filename(self) -> str:
        # TODO deprecate
        return str(self.path)

Result = Res[Book]

class Loader:
    def __init__(self, p: Path) -> None:
        self.path = p
        self.uid = self.path.parent.name

    def error(self, cause: Exception, extra: str ='') -> Exception:
        if len(extra) > 0:
            extra = '\n' + extra
        return echain(Exception(f'while processing {self.path}{extra}'), cause)

    def load_item(self, meta: Zoomable) -> Iterable[Highlight]:
        meta = cast(Wdict, meta)
        # TODO this should be destructive zoom?
        meta['notes'].zoom() # TODO ??? is it deliberate?

        meta['pagemarks'].consume_all()


        if 'notes' in meta:
            # TODO something nicer?
            notes = meta['notes'].zoom()
        else:
            notes = []
        comments = list(meta['comments'].zoom().values()) if 'comments' in meta else []
        meta['questions'].zoom()
        meta['flashcards'].zoom()
        highlights = meta['textHighlights'].zoom()

        # TODO could be useful to at least add a meta bout area highlights/screens
        meta['areaHighlights'].consume_all()
        meta['screenshots'].zoom()
        meta['thumbnails'].zoom()
        if 'readingProgress' in meta:
            meta['readingProgress'].consume_all()

        # TODO want to ignore the whole subtree..
        pi = meta['pageInfo'].zoom()
        page = pi['num'].zoom().value
        if 'dimensions' in pi:
            pi['dimensions'].consume_all()

        # TODO how to make it nicer?
        cmap: Dict[Hid, List[Comment]] = {}
        vals = list(comments)
        for v in vals:
            cid = v['id'].zoom()
            v['guid'].zoom()
            # TODO values should probably be checked by flow analysis??
            crt = v['created'].zoom()
            updated = v['lastUpdated'].zoom()
            content = v['content'].zoom()
            html = content['HTML'].zoom()
            refv = v['ref'].zoom().value
            [_, hlid] = refv.split(':')
            ccs = cmap.get(hlid, [])
            cmap[hlid] = ccs
            ccs.append(Comment(
                cid=cid.value,
                created=fromisoformat(crt.value),
                text=html.value, # TODO perhaps coonvert from html to text or org?
            ))
            v.consume()
        for h in list(highlights.values()):
            hid = h['id'].zoom().value
            if hid in cmap:
                comments = cmap[hid]
                del cmap[hid]
            else:
                comments = []

            h['guid'].consume()
            crt = h['created'].zoom().value
            updated = h['lastUpdated'].zoom().value
            h['rects'].ignore()

            # TODO make it more generic..
            htags: List[str] = []
            if 'tags' in h:
                ht = h['tags'].zoom()
                for k, v in list(ht.items()):
                    ctag = v.zoom()
                    ctag['id'].consume()
                    ct = ctag['label'].zoom()
                    htags.append(ct.value)

            h['textSelections'].ignore()
            h['notes'].consume()
            h['questions'].consume()
            h['flashcards'].consume()
            color = h['color'].zoom().value
            h['images'].ignore()
            # TODO eh, quite excessive \ns...
            text = h['text'].zoom()['TEXT'].zoom().value

            yield Highlight(
                hid=hid,
                created=fromisoformat(crt),
                selection=text,
                comments=tuple(comments),
                tags=tuple(htags),
                page=page,
                color=color,
            )
            h.consume()

        # TODO when I add defensive error policy, support it
        # if len(cmap) > 0:
        #     raise RuntimeError(f'Unconsumed comments: {cmap}')
        # TODO sort by date?


    def load_items(self, metas: Json) -> Iterable[Highlight]:
        for p, meta in metas.items():
            with wrap(meta, throw=not config.defensive) as meta:
                yield from self.load_item(meta)

    def load(self) -> Iterable[Result]:
        logger.info('processing %s', self.path)
        j = json.loads(self.path.read_text())

        # TODO konsume here as well?
        di = j['docInfo']
        added = di['added']
        filename = di['filename']
        title = di.get('title', None)
        tags_dict = di['tags']
        pm = j['pageMetas'] # todo handle this too?

        # todo defensive?
        tags = tuple(t['label'] for t in tags_dict.values())

        path = Path(config.polar_dir) / 'stash' / filename

        yield Book(
            created=fromisoformat(added),
            uid=self.uid,
            path=path,
            title=title,
            items=list(self.load_items(pm)),
            tags=tags,
        )


def iter_entries() -> Iterable[Result]:
    from .core import get_files
    for d in get_files(config.polar_dir, glob='*/state.json'):
        loader = Loader(d)
        try:
            yield from loader.load()
        except Exception as ee:
            err = loader.error(ee)
            logger.exception(err)
            yield err


def get_entries() -> List[Result]:
    # sorting by first annotation is reasonable I guess???
    # todo perhaps worth making it a pattern? X() returns iterable, get_X returns reasonably sorted list?
    return list(sort_res_by(iter_entries(), key=lambda e: e.created))


## deprecated
if not TYPE_CHECKING:
    # "deprecate" by hiding from mypy
    Error = Exception  # for backwards compat with Orger; can remove later



'''
Programmatic access and queries to org-mode files on the filesystem
'''

REQUIRES = [
    'orgparse',
]

from datetime import datetime
from pathlib import Path
import re
from typing import List, Sequence, Iterable, NamedTuple, Optional, Tuple

from my.core import get_files
from my.core.common import mcachew
from my.core.cachew import cache_dir
from my.core.orgmode import collect

from my.config import orgmode as user_config

import orgparse


# temporary? hack to cache org-mode notes
class OrgNote(NamedTuple):
    created: Optional[datetime]
    heading: str
    tags: List[str]


def inputs() -> Sequence[Path]:
    return get_files(user_config.paths)


_rgx = re.compile(orgparse.date.gene_timestamp_regex(brtype='inactive'), re.VERBOSE)
def _created(n: orgparse.OrgNode) -> Tuple[Optional[datetime], str]:
    heading = n.heading
    # meh.. support in orgparse?
    pp = {} if n.is_root() else n.properties
    createds = pp.get('CREATED', None)
    if createds is None:
        # try to guess from heading
        m = _rgx.search(heading)
        if m is not None:
            createds = m.group(0) # could be None
    if createds is None:
        return (None, heading)
    assert isinstance(createds, str)
    [odt] = orgparse.date.OrgDate.list_from_str(createds)
    dt = odt.start
    # todo a bit hacky..
    heading = heading.replace(createds + ' ', '')
    return (dt, heading)


def to_note(x: orgparse.OrgNode) -> OrgNote:
    # ugh. hack to merely make it cacheable
    heading = x.heading
    created: Optional[datetime]
    try:
        c, heading = _created(x)
        if isinstance(c, datetime):
            created = c
        else:
            # meh. not sure if should return date...
            created = None
    except Exception as e:
        created = None
    return OrgNote(
        created=created,
        heading=heading, # todo include the body?
        tags=list(x.tags),
    )


# todo move to common?
def _sanitize(p: Path) -> str:
    return re.sub(r'\W', '_', str(p))


def _cachew_cache_path(_self, f: Path) -> Path:
    return cache_dir() / 'orgmode' / _sanitize(f)


def _cachew_depends_on(_self, f: Path):
    return (f, f.stat().st_mtime)

 
class Query:
    def __init__(self, files: Sequence[Path]) -> None:
        self.files = files

    # TODO yield errors?
    @mcachew(
        cache_path=_cachew_cache_path, force_file=True,
        depends_on=_cachew_depends_on,
    )
    def _iterate(self, f: Path) -> Iterable[OrgNote]:
        o = orgparse.load(f)
        for x in o:
            yield to_note(x)

    def all(self) -> Iterable[OrgNote]:
        # TODO  build a virtual hierarchy from it?
        for f in self.files:
            yield from self._iterate(f)

    def collect_all(self, collector) -> Iterable[orgparse.OrgNode]:
        for f in self.files:
            o = orgparse.load(f)
            yield from collect(o, collector)


def query() -> Query:
    return Query(files=inputs())


from my.core import Stats, stat
def stats() -> Stats:
    def outlines():
        return query().all()
    return stat(outlines)



"""
An extension of the my.core.query.select function, allowing you to specify
a type or key to filter the range by -- this creates a filter function
given those values, coercing values on the iterable, returning you a
filtered iterator

See the select_range function below
"""

import re
import time
from functools import lru_cache
from datetime import datetime, timedelta, date
from typing import Callable, Iterator, NamedTuple, Optional, Any, Type

import more_itertools

from .query import (
    QueryException,
    select,
    OrderFunc,
    Where,
    _handle_generate_order_by,
    ET,
)

from .compat import fromisoformat


timedelta_regex = re.compile(r"^((?P<weeks>[\.\d]+?)w)?((?P<days>[\.\d]+?)d)?((?P<hours>[\.\d]+?)h)?((?P<minutes>[\.\d]+?)m)?((?P<seconds>[\.\d]+?)s)?$")


# https://stackoverflow.com/a/51916936
def parse_timedelta_string(timedelta_str: str) -> timedelta:
    """
    This uses a syntax similar to the 'GNU sleep' command
    e.g.: 1w5d5h10m50s means '1 week, 5 days, 5 hours, 10 minutes, 50 seconds'
    """
    parts = timedelta_regex.match(timedelta_str)
    if parts is None:
        raise ValueError(f"Could not parse time duration from {timedelta_str}.\nValid examples: '8h', '1w2d8h5m20s', '2m4s'")
    time_params = {name: float(param) for name, param in parts.groupdict().items() if param}
    return timedelta(**time_params)


def parse_timedelta_float(timedelta_str: str) -> float:
    return parse_timedelta_string(timedelta_str).total_seconds()


def parse_datetime_float(date_str: str) -> float:
    """
    parses multiple possible representations of a datetime
    into a float, else raises a QueryException

    the query_cli interface compares floats instead of timestamps
    when comparing datetimes since handling it is unknown
    whether the sources the user is selecting from is tz-aware
    or not (or perhaps a mix of both?)
    """
    ds = date_str.strip()
    # special case
    if ds == "now":
        return time.time()
    # epoch timestamp
    try:
        # also handles epoch timestamps as integers
        ds_float = float(ds)
        # convert to make sure its a valid datetime
        datetime.fromtimestamp(ds_float)
    except ValueError:
        pass
    else:
        return ds_float
    try:
        # isoformat - default format when you call str() on datetime
        # this also parses dates like '2020-01-01'
        return datetime.fromisoformat(ds).timestamp()
    except ValueError:
        pass
    try:
        return fromisoformat(ds).timestamp()
    except (AssertionError, ValueError):
        pass

    try:
        import dateparser
    except ImportError:
        pass
    else:
        # dateparser is a bit more lenient than the above, lets you type
        # all sorts of dates as inputs
        # https://github.com/scrapinghub/dateparser#how-to-use
        res: Optional[datetime] = dateparser.parse(ds, settings={"DATE_ORDER": "YMD"})
        if res is not None:
            return res.timestamp()

    raise QueryException(f"Was not able to parse {ds} into a datetime")


# probably DateLike input? but a user could specify an order_key
# which is an epoch timestamp or a float value which they
# expect to be converted to a datetime to compare
@lru_cache(maxsize=None)
def _datelike_to_float(dl: Any) -> float:
    if isinstance(dl, datetime):
        return dl.timestamp()
    elif isinstance(dl, date):
        # hmm... sets the hours/minutes/seconds to 0 -- make this configurable?
        return (datetime.combine(dl, datetime.min.time())).timestamp()
    else:
        try:
            return parse_datetime_float(dl)
        except QueryException as q:
            raise QueryException(f"While attempting to extract datetime from {dl}, to order by datetime:\n\n" + str(q))


class RangeTuple(NamedTuple):
    """Can specify 0, 1 or 2 non-none items in a range -- but not all 3

    As an example, using datetimes/timedelta (some date, and possibly a duration)

    where 1 arg is not None
        - after is not None: filters it to any items 'after' the datetime
        - before is not None: filters to any items 'before' the datetime
        - within: filters to any items 'within' the timedelta, assuming you meant within the current
            timeframe, so before = time.time()

    when 2 args are not None:
        - after and within, filters anything after the initial 'after' time
            but 'within' the timeframe (parsed timedelta, e.g. 5d)
        - before and within, anything 'within' the timeframe, starting at the end
            of the timeframe -- 'before'
        - before and after - anything after 'after' and before 'before', acts as a time range
    """
    # technically doesn't need to be Optional[Any],
    # just to make it more clear these can be None
    after: Optional[Any]
    before: Optional[Any]
    within: Optional[Any]


Converter = Callable[[Any], Any]


def _parse_range(
    *,
    unparsed_range: RangeTuple,
    end_parser: Converter,
    within_parser: Converter,
    parsed_range: Optional[RangeTuple] = None,
    error_message: Optional[str] = None
) -> Optional[RangeTuple]:

    if parsed_range is not None:
        return parsed_range

    err_msg = error_message or RangeTuple.__doc__
    assert err_msg is not None  # make mypy happy
    after, before, within = None, None, None

    none_count = more_itertools.ilen(filter(lambda o: o is None, list(unparsed_range)))
    if none_count == 3:
        return None
    if none_count == 0:
        raise QueryException("Cannot specify 'after', 'before' and 'within' at the same time!\n\n" + err_msg)

    [after_str, before_str, within_str] = tuple(unparsed_range)
    after = end_parser(after_str) if after_str is not None else None
    before = end_parser(before_str) if before_str is not None else None
    within = within_parser(within_str) if within_str is not None else None

    return RangeTuple(after=after, before=before, within=within)


def _create_range_filter(
    *,
    unparsed_range: RangeTuple,
    end_parser: Converter,
    within_parser: Converter,
    attr_func: Where,
    parsed_range: Optional[RangeTuple] = None,
    default_before: Optional[Any] = None,
    value_coercion_func: Optional[Converter] = None,
    error_message: Optional[str] = None,
) -> Optional[Where]:
    """
    Handles:
        - parsing the user input into values that are comparable to items the iterable returns
            - unparsed_range: tuple of raw values from user
            - end_parser: parses 'before' and 'after' (e.g. start/end dates)
            - within_parser: parser for the 'range' (e.g. timedelta)
            - error_message: allow overriding the default error message while parsing
        - converting items from the iterable to some coerced value, so that its comparable to
          the before, after and within parts of the range
            - if value_coercion_func is present, tries to use that
              to convert the value returned by the attr_func

    unparsed_range is a tuple of the input data from the user

    parsed_range can be passed if you've already parsed unparsed_range

    'default_before' specifies what to set if no before or after was specified in
    RangeTuple and we need an endpoint to end the range at. For example, if you wanted
    data from an iterable from the last week, you could specify default_before to be now (time.time()),
    and unparsed_range.within to be 7 days

    Creates a predicate that checks if some item from the iterator is
    within some range. this is typically used for datelike input, but the user could
    specify an integer or float item to order the values by/in some timeframe

    It requires the value you're comparing by to support comparable/addition operators (=, <, >, +, -)

    attr_func is a function which accepts the object from the iterator and returns
    the value to compare the range boundaries to. typically generated by _generate_order_by_func

    To force the values you're sorting by to be in some specified type,
    this allows a 'value_coercion_func', which optionally converts the value
    returned by attr_func to some shared type (see _datelike_to_float for an example)
    """

    rn = _parse_range(unparsed_range=unparsed_range,
                      end_parser=end_parser,
                      within_parser=within_parser,
                      parsed_range=parsed_range,
                      error_message=error_message)

    # user specified all 'None' items in the range, don't need to filter
    if rn is None:
        return None

    after = rn.after
    before = rn.before
    within = rn.within

    # hmm... not sure how to correctly manage
    # inclusivity here? Is [after, before) currently,
    # items are included on the lower bound but not the
    # upper bound
    # typically used for datetimes so doesn't have to
    # be exact in that case
    def generated_predicate(obj: Any) -> bool:
        ov: Any = attr_func(obj)
        if value_coercion_func is not None:
            ov = value_coercion_func(ov)
        if after is not None:
            if before is not None:
                # squeeze between before/after
                return ov >= after and ov < before
            elif within is not None:
                # after some start point + some range
                allow_before = after + within
                return ov >= after and ov < allow_before
            else:
                return ov >= after
        elif before is not None:
            if within is not None:
                allow_after = before - within
                # before a startpoint + some range
                return ov >= allow_after and ov < before
            else:
                # just before the startpoint
                return ov < before
        else:
            # only specified within, default before to now
            if default_before is None:
                raise QueryException("Only received a range length, with no start or end point to compare against")
            allow_after = default_before - within
            return ov >= allow_after and ov < default_before

    return generated_predicate


# main interface to this file from my.core.__main__.py
def select_range(
    itr: Iterator[ET],
    *,
    where: Optional[Where] = None,
    order_key: Optional[str] = None,
    order_value: Optional[Where] = None,
    order_by_value_type: Optional[Type] = None,
    unparsed_range: Optional[RangeTuple] = None,
    reverse: bool = False,
    limit: Optional[int] = None,
    drop_unsorted: bool = False,
    wrap_unsorted: bool = False,
    warn_exceptions: bool = False,
    warn_func: Optional[Callable[[Exception], None]] = None,
    drop_exceptions: bool = False,
    raise_exceptions: bool = False,
) -> Iterator[ET]:
    """
    A specialized select function which offers generating functions
    to filter/query ranges from an iterable

    order_key and order_value are used in the same way they are in select

    If you specify order_by_value_type, it tries to search for an attribute
    on each object/type which has that type, ordering the iterable by that value

    unparsed_range is a tuple of length 3, specifying 'after', 'before', 'duration',
    i.e. some start point to allow the computed value we're ordering by, some
    end point and a duration (can use the RangeTuple NamedTuple to construct one)

    (this is typically parsed/created in my.core.__main__, from CLI flags

    If you specify a range, drop_unsorted is forced to be True
    """

    # if the user specified a range with no data, set the unparsed_range to None
    if unparsed_range == RangeTuple(None, None, None):
        unparsed_range = None

    # some operations to do before ordering/filtering
    if drop_exceptions or raise_exceptions or where is not None or warn_exceptions:
        # doesn't wrap unsortable items, because we pass no order related kwargs
        itr = select(
            itr,
            where=where,
            drop_exceptions=drop_exceptions,
            raise_exceptions=raise_exceptions,
            warn_exceptions=warn_exceptions,
            warn_func=warn_func)

    order_by_chosen: Optional[OrderFunc] = None

    # if the user didn't specify an attribute to order value, but specified a type
    # we should search for on each value in the iterator
    if order_value is None and order_by_value_type is not None:
        # search for that type on the iterator object
        order_value = lambda o: isinstance(o, order_by_value_type)

    # if the user supplied a order_key, and/or we've generated an order_value, create
    # the function that accesses that type on each value in the iterator
    if order_key is not None or order_value is not None:
        order_by_chosen, itr = _handle_generate_order_by(itr, order_key=order_key, order_value=order_value)
        # signifies that itr is empty -- can early return here
        if order_by_chosen is None:
            return itr

    # test if the user is trying to specify a range to filter the items by
    if unparsed_range is not None:

        if order_by_chosen is None:
            raise QueryException("""Can't order by range if we have no way to order_by!
Specify a type or a key to order the value by""")
        else:
            # force drop_unsorted=True so we can use _create_range_filter
            # sort the iterable by the generated order_by_chosen function
            itr = select(itr, order_by=order_by_chosen, drop_unsorted=True)
            filter_func: Optional[Where]
            if order_by_value_type in [datetime, date]:
                filter_func = _create_range_filter(
                    unparsed_range=unparsed_range,
                    end_parser=parse_datetime_float,
                    within_parser=parse_timedelta_float,
                    attr_func=order_by_chosen,  # type: ignore[arg-type]
                    default_before=time.time(),
                    value_coercion_func=_datelike_to_float)
            elif order_by_value_type in [int, float]:
                # allow primitives to be converted using the default int(), float() callables
                filter_func = _create_range_filter(
                    unparsed_range=unparsed_range,
                    end_parser=order_by_value_type,
                    within_parser=order_by_value_type,
                    attr_func=order_by_chosen,  # type: ignore[arg-type]
                    default_before=None,
                    value_coercion_func=order_by_value_type)
            else:
                # TODO: add additional kwargs to let the user sort by other values, by specifying the parsers?
                # would need to allow passing the end_parser, within parser, default before and value_coercion_func...
                # (seems like a lot?)
                raise QueryException("Sorting by custom types is currently unsupported")

            # use the created filter function
            # we've already applied drop_exceptions and kwargs related to unsortable values above
            itr = select(itr, where=filter_func, limit=limit, reverse=reverse)
    else:
        # wrap_unsorted may be used here if the user specified an order_key,
        # or manually passed a order_value function
        #
        # this select is also run if the user didn't specify anything to
        # order by, and is just returning the data in the same order as
        # as the source iterable
        # i.e. none of the range-related filtering code ran, this is just a select
        itr = select(itr,
                     order_by=order_by_chosen,
                     wrap_unsorted=wrap_unsorted,
                     drop_unsorted=drop_unsorted,
                     limit=limit,
                     reverse=reverse)
    return itr


# re-use items from query for testing
from .query import _A, _B, _Float, _mixed_iter_errors


def test_filter_in_timeframe() -> None:

    from itertools import chain

    jan_1_2005 = datetime(year=2005, month=1, day=1, hour=1, minute=1, second=1)
    jan_1_2016 = datetime(year=2016, month=1, day=1, hour=1, minute=1, second=1)

    rng = RangeTuple(after=str(jan_1_2005), before=str(jan_1_2016), within=None)

    # items between 2005 and 2016
    res = list(select_range(_mixed_iter_errors(), order_by_value_type=datetime, unparsed_range=rng, drop_exceptions=True))

    assert res == [_A(x=datetime(2005, 4, 10, 4, 10, 1), y=2, z=-5),
                   _A(x=datetime(2005, 5, 10, 4, 10, 1), y=10, z=2),
                   _A(x=datetime(2009, 3, 10, 4, 10, 1), y=12, z=1),
                   _A(x=datetime(2009, 5, 10, 4, 10, 1), y=5, z=10),
                   _B(y=datetime(year=2015, month=5, day=10, hour=4, minute=10, second=1))]

    rng = RangeTuple(before=str(jan_1_2016), within="52w", after=None)

    # from 2016, going back 52 weeks (about a year?)
    res = list(select_range(_mixed_iter_errors(), order_by_value_type=datetime, unparsed_range=rng, drop_exceptions=True))

    assert res == [_B(y=datetime(year=2015, month=5, day=10, hour=4, minute=10, second=1))]

    # test passing just a within while using a datetime. should default to using current time
    recent_time = datetime.now() - timedelta(days=5)
    obj = _A(x=recent_time, y=2, z=-5)

    rng = RangeTuple(before=None, after=None, within="1w")
    res = list(select_range(chain(_mixed_iter_errors(), iter([obj])),
                            order_by_value_type=datetime,
                            unparsed_range=rng, drop_exceptions=True))

    assert res == [obj]

    # dont pass any range related stuff, use where/drop_exceptions and the limit flag
    # to make sure this falls through properly to using select kwargs

    using_range = list(select_range(_mixed_iter_errors(), drop_exceptions=True, limit=5))
    normal = list(select(_mixed_iter_errors(), limit=5, where=lambda o: not isinstance(o, Exception)))

    assert using_range == normal


def test_query_range_float_value_type() -> None:

    def floaty_iter() -> Iterator[_Float]:
        for v in range(1, 6):
            yield _Float(float(v + 0.5))

    rng = RangeTuple(after=2, before=6.1, within=None)
    res = list(select_range(floaty_iter(), order_by_value_type=float, unparsed_range=rng, drop_exceptions=True))
    assert res == [_Float(2.5), _Float(3.5), _Float(4.5), _Float(5.5)]


def test_range_predicate() -> None:

    from functools import partial

    def src() -> Iterator[str]:
        yield from map(str, range(15))

    identity = lambda o: o

    # convert any float values to ints
    coerce_int_parser = lambda o: int(float(o))
    int_filter_func = partial(
        _create_range_filter,
        attr_func=identity,
        end_parser=coerce_int_parser,
        within_parser=coerce_int_parser,
        value_coercion_func=coerce_int_parser,
    )

    # filter from 0 to 5
    rn: Optional[RangeTuple] = RangeTuple("0", "5", None)
    zero_to_five_filter: Optional[Where] = int_filter_func(unparsed_range=rn)
    assert zero_to_five_filter is not None
    # this is just a Where function, given some input it return True/False if the value is allowed
    assert zero_to_five_filter(3) is True
    assert zero_to_five_filter(10) is False

    # this is expected, range_predicate is not inclusive on the far end
    assert list(filter(zero_to_five_filter, src())) == ["0", "1", "2", "3", "4"]

    # items less than 3, going 3.5 (converted to 3 by the coerce_int_parser) down
    rn = RangeTuple(None, 3, "3.5")
    assert list(filter(int_filter_func(unparsed_range=rn, attr_func=identity), src())) == ["0", "1", "2"]

def test_parse_range() -> None:

    from functools import partial

    import pytest

    rn = RangeTuple("0", "5", None)
    res = _parse_range(unparsed_range=rn, end_parser=int, within_parser=int)

    assert res == RangeTuple(after=0, before=5, within=None)

    dt_parse_range = partial(_parse_range, end_parser=parse_datetime_float, within_parser=parse_timedelta_float)

    start_date = datetime.now()
    end_date = start_date + timedelta(seconds=60)

    # convert start items to strings, which need to be parsed back
    rn = RangeTuple(str(start_date), str(end_date.timestamp()), None)
    res2 = dt_parse_range(unparsed_range=rn)

    assert res2 == RangeTuple(after=start_date.timestamp(), before=end_date.timestamp(), within=None)

    # can't specify all three
    with pytest.raises(QueryException, match=r"Cannot specify 'after', 'before' and 'within'"):
        dt_parse_range(unparsed_range=RangeTuple(str(start_date), str(end_date.timestamp()), "7d"))

    # if you specify noting, should return None
    res3 = dt_parse_range(unparsed_range=RangeTuple(None, None, None))
    assert res3 is None


def test_parse_timedelta_string() -> None:

    import pytest

    with pytest.raises(ValueError, match=r"Could not parse time duration from"):
        parse_timedelta_string("5xxx")

    res = parse_timedelta_string("1w5d5h10m50s")
    assert res == timedelta(days=7.0 + 5.0, hours=5.0, minutes=10.0, seconds=50.0)


def test_parse_datetime_float() -> None:

    pnow = parse_datetime_float("now")
    sec_diff = abs((pnow - datetime.now().timestamp()))
    # should probably never fail? could mock time.time
    # but there seems to be issues with doing that use C-libraries (as time.time) does
    # https://docs.python.org/3/library/unittest.mock-examples.html#partial-mocking
    assert sec_diff < 60

    dt = datetime.now()
    dt_float_s = str(dt.timestamp())
    dt_int_s = str(int(dt.timestamp()))

    # float/int representations as strings
    assert dt.timestamp() == parse_datetime_float(dt_float_s)
    assert int(dt.timestamp()) == int(parse_datetime_float(dt_int_s))

    # test parsing isoformat
    assert dt.timestamp() == parse_datetime_float(str(dt))



'''
Rescuetime (phone activity tracking) data.
'''
REQUIRES = [
    'git+https://github.com/karlicoss/rescuexport',
]

from pathlib import Path
from datetime import timedelta
from typing import Sequence, Iterable

from my.core import get_files, make_logger
from my.core.common import mcachew
from my.core.error import Res, split_errors

from my.config import rescuetime as config


logger = make_logger(__name__)


def inputs() -> Sequence[Path]:
    return get_files(config.export_path)


import rescuexport.dal as dal
DAL = dal.DAL
Entry = dal.Entry


@mcachew(depends_on=inputs)
def entries() -> Iterable[Res[Entry]]:
    dal = DAL(inputs())
    yield from dal.entries()


def groups(gap: timedelta=timedelta(hours=3)) -> Iterable[Res[Sequence[Entry]]]:
    vit, eit = split_errors(entries(), ET=Exception)
    yield from eit
    from more_itertools import split_when
    yield from split_when(vit, lambda a, b: (b.dt - a.dt) > gap)


# todo automatic dataframe interface?
from .core.pandas import DataFrameT, as_dataframe
def dataframe() -> DataFrameT:
    return as_dataframe(entries())


from .core import stat, Stats
def stats() -> Stats:
    return {
        **stat(groups),
        **stat(entries),
    }


# basically, hack config and populate it with fake data? fake data generated by DAL, but the rest is handled by this?

from contextlib import contextmanager
from typing import Iterator
# todo take seed, or what?
@contextmanager
def fake_data(rows: int=1000) -> Iterator:
    # todo also disable cachew automatically for such things?
    from my.core.cfg import tmp_config
    from my.core.cachew import disabled_cachew
    from tempfile import TemporaryDirectory
    import json
    with disabled_cachew(), TemporaryDirectory() as td:
        tdir = Path(td)
        f = tdir / 'rescuetime.json'
        f.write_text(json.dumps(dal.fake_data_generator(rows=rows)))

        class override:
            class rescuetime:
                export_path = tdir

        with tmp_config(modules=__name__, config=override) as cfg:
            yield cfg
# TODO ok, now it's something that actually could run on CI!
# todo would be kinda nice if doctor could run against the fake data, to have a basic health check of the module?


def fill_influxdb() -> None:
    from .core import influxdb
    it = (dict(
        dt=e.dt,
        duration_d=e.duration_s,
        tags=dict(activity=e.activity),
    ) for e in entries() if isinstance(e, Entry)) # TODO handle errors in core.influxdb
    influxdb.fill(it, measurement=__name__)


# TODO lots of garbage in dir()? maybe need to del the imports...



from my.tests.common import skip_if_not_karlicoss as pytestmark

# TODO move elsewhere?
# these tests would only make sense with some existing data? although some of them would work for everyone..
# not sure what's a good way of handling this..
from datetime import datetime
import pytz
from more_itertools import bucket


# TODO ugh. if i uncomment this here (on top level), then this test vvv fails
# from my.media.youtube import get_watched, Watched
# HPI_TESTS_KARLICOSS=true pytest -raps tests/tz.py tests/youtube.py


def test() -> None:
    from my.youtube.takeout import watched, Watched
    videos = [w for w in watched() if not isinstance(w, Exception)]
    assert len(videos) > 1000

    # results in nicer errors, otherwise annoying to check against thousands of videos
    grouped = bucket(videos, key=lambda w: (w.url, w.title))

    w1 = Watched(
        url='https://www.youtube.com/watch?v=hTGJfRPLe08',
        title='Jamie xx - Gosh',
        when=pytz.timezone('Europe/London').localize(datetime(year=2018, month=6, day=21, hour=6, minute=48, second=34)),
    )
    assert w1 in list(grouped[(w1.url, w1.title)])

    w2 = Watched(
        url='https://www.youtube.com/watch?v=IZ_8b_Ydsv0',
        title='Why LESS Sensitive Tests Might Be Better',
        when=pytz.utc.localize(datetime(year=2021, month=1, day=15, hour=17, minute=54, second=12)),
    )
    assert w2 in list(grouped[(w2.url, w2.title)])



from my.core.cfg import tmp_config
from my.core.common import make_dict

# todo ugh, it's discovered as a test???
from .common import testdata

import pytest

# deliberately use mixed style imports on the top level and inside the methods to test tmp_config stuff
import my.reddit.rexport as my_reddit_rexport
import my.reddit.all as my_reddit_all


def test_basic_1() -> None:
    # todo maybe this should call stat or something instead?
    # would ensure reasonable stat implementation as well and less duplication
    # note: deliberately use old module (instead of my.reddit.all) to test bwd compatibility
    from my.reddit import saved

    assert len(list(saved())) > 0


def test_basic_2() -> None:
    # deliberately check call from a different style of import to make sure tmp_config works
    saves = list(my_reddit_rexport.saved())
    assert len(saves) > 0


def test_comments() -> None:
    assert len(list(my_reddit_all.comments())) > 0


def test_saves() -> None:
    from my.reddit.all import saved

    saves = list(saved())
    assert len(saves) > 0

    # just check that they are unique (makedict will throw)
    make_dict(saves, key=lambda s: s.sid)


def test_preserves_extra_attr() -> None:
    # doesn't strictly belong here (not specific to reddit)
    # but my.reddit does a fair bit of dynamic hacking, so perhaps a good place to check nothing is lost
    from my.reddit import config

    assert isinstance(getattr(config, 'please_keep_me'), str)


@pytest.fixture(autouse=True, scope='module')
def prepare():
    data = testdata() / 'hpi-testdata' / 'reddit'
    assert data.exists(), data

    # note: deliberately using old config schema so we can test migrations
    class config:
        class reddit:
            export_dir = data
            please_keep_me = 'whatever'

    with tmp_config(modules='my.reddit.*', config=config):
        yield



from datetime import datetime
from typing import Callable, cast

from ...core.common import tzdatetime, Literal


'''
Depending on the specific data provider and your level of paranoia you might expect different behaviour.. E.g.:
- if your objects already have tz info, you might not need to call localize() at all
- it's safer when either all of your objects are tz aware or all are tz unware, not a mixture
- you might trust your original timezone, or it might just be UTC, and you want to use something more reasonable
'''
TzPolicy = Literal[
    'keep'   , # if datetime is tz aware, just preserve it
    'convert', # if datetime is tz aware, convert to provider's tz
    'throw'  , # if datetime is tz aware, throw exception
    # todo 'warn'? not sure if very useful
]

# backwards compatibility
Policy = TzPolicy

def default_policy() -> TzPolicy:
    try:
        from my.config import time as user_config
        return cast(TzPolicy, user_config.tz.policy)
    except Exception as e:
        # todo meh.. need to think how to do this more carefully
        # rationale: do not mess with user's data unless they want
        return 'keep'


def localize_with_policy(lfun: Callable[[datetime], tzdatetime], dt: datetime, policy: TzPolicy=default_policy()) -> tzdatetime:
    tz = dt.tzinfo
    if tz is None:
        return lfun(dt)

    if   policy == 'keep':
        return dt
    elif policy == 'convert':
        ldt = lfun(dt.replace(tzinfo=None))
        return dt.astimezone(ldt.tzinfo)
    else: # policy == 'error':
        raise RuntimeError(f"{dt} already has timezone information (use 'policy' argument to adjust this behaviour)")



"""
Various error handling helpers
See https://beepb00p.xyz/mypy-error-handling.html#kiss for more detail
"""

from itertools import tee
from typing import Union, TypeVar, Iterable, List, Tuple, Type, Optional, Callable, Any, cast, Iterator, Literal


T = TypeVar('T')
E = TypeVar('E', bound=Exception) # TODO make covariant?

ResT = Union[T, E]

Res = ResT[T, Exception]

ErrorPolicy = Literal["yield", "raise", "drop"]

def notnone(x: Optional[T]) -> T:
    assert x is not None
    return x


def unwrap(res: Res[T]) -> T:
    if isinstance(res, Exception):
        raise res
    else:
        return res

def drop_exceptions(itr: Iterator[Res[T]]) -> Iterator[T]:
    """Return non-errors from the iterable"""
    for o in itr:
        if isinstance(o, Exception):
            continue
        yield o


def raise_exceptions(itr: Iterable[Res[T]]) -> Iterator[T]:
    """Raise errors from the iterable, stops the select function"""
    for o in itr:
        if isinstance(o, Exception):
            raise o
        yield o


def warn_exceptions(itr: Iterable[Res[T]], warn_func: Optional[Callable[[Exception], None]] = None) -> Iterator[T]:
    # if not provided, use the 'warnings' module
    if warn_func is None:
        from my.core.warnings import medium
        def _warn_func(e: Exception) -> None:
            # TODO: print traceback? but user could always --raise-exceptions as well
            medium(str(e))
        warn_func = _warn_func

    for o in itr:
        if isinstance(o, Exception):
            warn_func(o)
            continue
        yield o


def echain(ex: E, cause: Exception) -> E:
    ex.__cause__ = cause
    return ex


def split_errors(l: Iterable[ResT[T, E]], ET: Type[E]) -> Tuple[Iterable[T], Iterable[E]]:
    # TODO would be nice to have ET=Exception default? but it causes some mypy complaints?
    vit, eit = tee(l)
    # TODO ugh, not sure if I can reconcile type checking and runtime and convince mypy that ET and E are the same type?
    values: Iterable[T] = (
        r # type: ignore[misc]
        for r in vit
        if not isinstance(r, ET))
    errors: Iterable[E] = (
        r
        for r in eit
        if     isinstance(r, ET))
    # TODO would be interesting to be able to have yield statement anywehere in code
    # so there are multiple 'entry points' to the return value
    return (values, errors)


K = TypeVar('K')
def sort_res_by(items: Iterable[Res[T]], key: Callable[[Any], K]) -> List[Res[T]]:
    """
    Sort a sequence potentially interleaved with errors/entries on which the key can't be computed.
    The general idea is: the error sticks to the non-error entry that follows it
    """
    group = []
    groups = []
    for i in items:
        k: Optional[K]
        try:
            k = key(i)
        except Exception:  # error white computing key? dunno, might be nice to handle...
            k = None
        group.append(i)
        if k is not None:
            groups.append((k, group))
            group = []

    results: List[Res[T]] = []
    for v, grp in sorted(groups, key=lambda p: p[0]): # type: ignore[return-value, arg-type] # TODO SupportsLessThan??
        results.extend(grp)
    results.extend(group) # handle last group (it will always be errors only)

    return results


def test_sort_res_by() -> None:
    class Exc(Exception):
        def __eq__(self, other):
            return self.args == other.args

    ress = [
        Exc('first'),
        Exc('second'),
        5,
        3,
        'bad',
        2,
        1,
        Exc('last'),
    ]
    results = sort_res_by(ress, lambda x: int(x))
    assert results == [
        1,
        'bad',
        2,
        3,
        Exc('first'),
        Exc('second'),
        5,
        Exc('last'),
    ]

    results2 = sort_res_by(ress + [0], lambda x: int(x))
    assert results2 == [Exc('last'), 0] + results[:-1]

    assert sort_res_by(['caba', 'a', 'aba', 'daba'], key=lambda x: len(x)) == ['a', 'aba', 'caba', 'daba']
    assert sort_res_by([], key=lambda x: x) == []


# helpers to associate timestamps with the errors (so something meaningful could be displayed on the plots, for example)
# todo document it under 'patterns' somewhere...

# todo proper typevar?
from datetime import datetime
def set_error_datetime(e: Exception, dt: Optional[datetime]) -> None:
    if dt is None:
        return
    e.args = e.args + (dt,)
    # todo not sure if should return new exception?

def attach_dt(e: Exception, *, dt: Optional[datetime]) -> Exception:
    set_error_datetime(e, dt)
    return e

# todo it might be problematic because might mess with timezones (when it's converted to string, it's converted to a shift)
def extract_error_datetime(e: Exception) -> Optional[datetime]:
    import re
    from datetime import datetime
    for x in reversed(e.args):
        if isinstance(x, datetime):
            return x
        if not isinstance(x, str):
            continue
        m = re.search(r'\d{4}-\d\d-\d\d(...:..:..)?(\.\d{6})?(\+.....)?', x)
        if m is None:
            continue
        ss = m.group(0)
        # todo not sure if should be defensive??
        return datetime.fromisoformat(ss)
    return None


import traceback
from .common import Json
def error_to_json(e: Exception) -> Json:
    estr = ''.join(traceback.format_exception(Exception, e, e.__traceback__))
    return {'error': estr}


MODULE_SETUP_URL = 'https://github.com/karlicoss/HPI/blob/master/doc/SETUP.org#private-configuration-myconfig'

def warn_my_config_import_error(err: Union[ImportError, AttributeError], help_url: Optional[str] = None) -> bool:
    """
    If the user tried to import something from my.config but it failed,
    possibly due to missing the config block in my.config?

    Returns True if it matched a possible config error
    """
    import re
    import click
    if help_url is None:
        help_url = MODULE_SETUP_URL
    if type(err) == ImportError:
        if err.name != 'my.config':
            return False
        # parse name that user attempted to import
        em = re.match(r"cannot import name '(\w+)' from 'my.config'", str(err))
        if em is not None:
            section_name = em.group(1)
            click.secho(f"""\
You may be missing the '{section_name}' section from your config.
See {help_url}\
""", fg='yellow', err=True)
            return True
    elif type(err) == AttributeError:
        # test if user had a nested config block missing
        # https://github.com/karlicoss/HPI/issues/223
        if hasattr(err, 'obj') and hasattr(err, "name"):
            config_obj = cast(object, getattr(err, 'obj'))  # the object that caused the attribute error
            # e.g. active_browser for my.browser
            nested_block_name = err.name
            if config_obj.__module__ == 'my.config':
                click.secho(f"""You're likely missing the nested config block for '{getattr(config_obj, '__name__', str(config_obj))}.{nested_block_name}'.
See {help_url} or check the corresponding module.py file for an example\
""", fg='yellow', err=True)
                return True
    else:
        click.echo(f"Unexpected error... {err}", err=True)
    return False


def test_datetime_errors() -> None:
    import pytz
    dt_notz = datetime.now()
    dt_tz   = datetime.now(tz=pytz.timezone('Europe/Amsterdam'))
    for dt in [dt_tz, dt_notz]:
        e1 = RuntimeError('whatever')
        assert extract_error_datetime(e1) is None
        set_error_datetime(e1, dt=dt)
        assert extract_error_datetime(e1) == dt

        e2 = RuntimeError(f'something something {dt} something else')
        assert extract_error_datetime(e2) == dt

    e3 = RuntimeError(str(['one', '2019-11-27T08:56:00', 'three']))
    assert extract_error_datetime(e3) is not None

    # date only
    e4 = RuntimeError(str(['one', '2019-11-27', 'three']))
    assert extract_error_datetime(e4) is not None



"""
Helper 'module' for test_guess_stats
"""
from contextlib import contextmanager
from dataclasses import dataclass
from datetime import datetime, timedelta
from pathlib import Path
from typing import Iterable, Sequence, Iterator


@dataclass
class Item:
    id: str
    dt: datetime
    source: Path


def inputs() -> Sequence[Path]:
    return [
        Path('file1.json'),
        Path('file2.json'),
        Path('file3.json'),
    ]


def iter_data() -> Iterable[Item]:
    dt = datetime.fromisoformat('2020-01-01 01:01:01')
    for path in inputs():
        for i in range(3):
            yield Item(id=str(i), dt=dt + timedelta(days=i), source=path)


@contextmanager
def some_contextmanager() -> Iterator[str]:
    # this shouldn't end up in guess_stats because context manager is not a data provider
    yield 'hello'



# TODO: add config here which passes kwargs to estimate_from (under_accuracy)
# overwritable by passing the kwarg name here to the top-level estimate_location

from typing import Iterator, Optional

from my.core.source import import_source
from my.location.fallback.common import (
    estimate_from,
    FallbackLocation,
    DateExact,
    LocationEstimator,
)


def fallback_locations() -> Iterator[FallbackLocation]:
    # can comment/uncomment sources here to enable/disable them
    yield from _ip_fallback_locations()


def fallback_estimators() -> Iterator[LocationEstimator]:
    # can comment/uncomment estimators here to enable/disable them
    # the order of the estimators determines priority if location accuries are equal/unavailable
    yield _ip_estimate
    yield _home_estimate


def estimate_location(dt: DateExact, first_match: bool=False, under_accuracy: Optional[int] = None) -> FallbackLocation:
    loc = estimate_from(dt, estimators=list(fallback_estimators()), first_match=first_match, under_accuracy=under_accuracy)
    # should never happen if the user has home configured
    if loc is None:
        raise ValueError("Could not estimate location")
    return loc


@import_source(module_name="my.location.fallback.via_home")
def _home_estimate(dt: DateExact) -> Iterator[FallbackLocation]:
    from my.location.fallback.via_home import estimate_location as via_home_estimate

    yield from via_home_estimate(dt)


@import_source(module_name="my.location.fallback.via_ip")
def _ip_estimate(dt: DateExact) -> Iterator[FallbackLocation]:
    from my.location.fallback.via_ip import estimate_location as via_ip_estimate

    yield from via_ip_estimate(dt)


@import_source(module_name="my.location.fallback.via_ip")
def _ip_fallback_locations() -> Iterator[FallbackLocation]:
    from my.location.fallback.via_ip import fallback_locations as via_ip_fallback

    yield from via_ip_fallback()



"""
[[https://www.instapaper.com][Instapaper]] bookmarks, highlights and annotations
"""
REQUIRES = [
    'git+https://github.com/karlicoss/instapexport',
]

from dataclasses import dataclass

from .core import Paths

from my.config import instapaper as user_config


@dataclass
class instapaper(user_config):
    '''
    Uses [[https://github.com/karlicoss/instapexport][instapexport]] outputs.
    '''
    # path[s]/glob to the exported JSON data
    export_path : Paths


from .core.cfg import make_config
config = make_config(instapaper)


try:
    from instapexport import dal
except ModuleNotFoundError as e:
    from my.core.hpi_compat import pre_pip_dal_handler

    dal = pre_pip_dal_handler('instapexport', e, config, requires=REQUIRES)

############################

Highlight = dal.Highlight
Bookmark  = dal.Bookmark
Page      = dal.Page


from typing import Sequence, Iterable
from pathlib import Path
from .core import get_files
def inputs() -> Sequence[Path]:
    return get_files(config.export_path)


def _dal() -> dal.DAL:
    return dal.DAL(inputs())


def pages() -> Iterable[Page]:
    return _dal().pages()


def stats():
    from .core import stat
    return stat(pages)

### misc

# TODO dunno, move this to private?
def is_todo(hl: Highlight) -> bool:
    note = hl.note or ''
    note = note.lstrip().lower()
    return note.startswith('todo')


get_pages = pages # todo also deprecate..



from pathlib import Path
import sys
from importlib import reload
from my.core.common import get_valid_filename

ROOT = Path(__file__).parent.absolute()
OUTPUTS = ROOT / 'outputs'


import pytest


def test_hpi(prepare: str) -> None:
    from my.polar import get_entries
    assert len(list(get_entries())) > 1

def test_orger(prepare: str, tmp_path: Path) -> None:
    from my.core.common import import_from, import_file
    om = import_file(ROOT / 'orger/modules/polar.py')
    # reload(om)

    pv = om.PolarView()
    # TODO hmm. worth making public?
    OUTPUTS.mkdir(exist_ok=True)
    out = OUTPUTS / (get_valid_filename(prepare) + '.org')
    pv._run(to=out)


PARAMS = [
    # 'data/polar/BojanKV_polar/.polar',
    '',
    # 'data/polar/TheCedarPrince_KnowledgeRepository',
    # 'data/polar/coelias_polardocs',
    # 'data/polar/warkdarrior_polar-document-repository'
]

@pytest.fixture(params=PARAMS)
def prepare(request):
    dotpolar = request.param
    class user_config:
        if dotpolar != '': # default
            polar_dir = Path(ROOT / dotpolar)
        defensive = False

    import my.config
    setattr(my.config, 'polar', user_config)

    import my.polar as polar
    reload(polar)
    # TODO hmm... ok, need to document reload()
    yield dotpolar



"""
Extracts semantic location history using google_takeout_parser
"""

# This is a separate module to prevent ImportError and a new config block from breaking
# previously functional my.location.google_takeout locations

REQUIRES = ["git+https://github.com/seanbreckenridge/google_takeout_parser"]

from typing import Iterator, List

from my.google.takeout.parser import events, _cachew_depends_on as _parser_cachew_depends_on
from google_takeout_parser.models import PlaceVisit as SemanticLocation

from my.core import dataclass, make_config
from my.core.common import mcachew, LazyLogger, Stats
from my.core.error import Res
from .common import Location

logger = LazyLogger(__name__)

from my.config import location as user_config

@dataclass
class semantic_locations_config(user_config.google_takeout_semantic):
    # a value between 0 and 100, 100 being the most confident
    # set to 0 to include all locations
    # https://locationhistoryformat.com/reference/semantic/#/$defs/placeVisit/properties/locationConfidence
    require_confidence: int = 40
    # default accuracy for semantic locations
    accuracy: float = 100.0


config = make_config(semantic_locations_config)


# add config to cachew dependency so it recomputes on config changes
def _cachew_depends_on() -> List[str]:
    dep = _parser_cachew_depends_on()
    dep.insert(0, f"require_confidence={config.require_confidence} accuracy={config.accuracy}")
    return dep



@mcachew(
    depends_on=_cachew_depends_on,
    logger=logger,
)
def locations() -> Iterator[Res[Location]]:
    require_confidence = config.require_confidence
    if require_confidence < 0 or require_confidence > 100:
        yield ValueError("location.google_takeout.semantic_require_confidence must be between 0 and 100")
        return

    for g in events():
        if isinstance(g, SemanticLocation):
            visitConfidence = g.visitConfidence
            if visitConfidence is None or visitConfidence < require_confidence:
                logger.debug(f"Skipping {g} due to low confidence ({visitConfidence}))")
                continue
            yield Location(
                lon=g.lng,
                lat=g.lat,
                dt=g.dt,
                # can accuracy be inferred from visitConfidence?
                # there's no exact distance value in the data, its a 0-100% confidence value...
                accuracy=config.accuracy,
                elevation=None,
                datasource="google_takeout_semantic",
            )



def stats() -> Stats:
    from my.core import stat

    return {**stat(locations)}



from pathlib import Path
from typing import Dict

import PIL.Image
from PIL.ExifTags import TAGS, GPSTAGS


Exif = Dict

# TODO PIL.ExifTags.TAGS


class ExifTags:
    DATETIME = "DateTimeOriginal"
    LAT      = "GPSLatitude"
    LAT_REF  = "GPSLatitudeRef"
    LON      = "GPSLongitude"
    LON_REF  = "GPSLongitudeRef"
    GPSINFO  = "GPSInfo"


# TODO there must be something more standard for this...
def get_exif_from_file(path: Path) -> Exif:
    # TODO exception handler?
    with PIL.Image.open(str(path)) as fo:
        return _get_exif_data(fo)


def _get_exif_data(image) -> Exif:
    """Returns a dictionary from the exif data of an PIL Image item. Also converts the GPS Tags"""
    exif_data = {}
    info = image._getexif()
    if info:
        for tag, value in info.items():
            decoded = TAGS.get(tag, tag)
            if decoded == ExifTags.GPSINFO:
                gps_data = {}
                for t in value:
                    sub_decoded = GPSTAGS.get(t, t)
                    gps_data[sub_decoded] = value[t]

                exif_data[decoded] = gps_data
            else:
                exif_data[decoded] = value

    return exif_data


def to_degree(value) -> float:
    """Helper function to convert the GPS coordinates
    stored in the EXIF to digress in float format"""
    (d, m, s) = value
    return d + (m / 60.0) + (s / 3600.0)


def convert_ref(cstr, ref: str) -> float:
    val = to_degree(cstr)
    if ref == 'S' or ref == 'W':
        val = -val
    return val


import re
from datetime import datetime
from typing import Optional

# TODO surely there is a library that does it??
# TODO this belongs to a private overlay or something
# basically have a function that patches up dates after the files were yielded..
_DT_REGEX = re.compile(r'\D(\d{8})\D*(\d{6})\D')
def dt_from_path(p: Path) -> Optional[datetime]:
    name = p.stem
    mm = _DT_REGEX.search(name)
    if mm is None:
        return None
    dates = mm.group(1) + mm.group(2)
    return datetime.strptime(dates, "%Y%m%d%H%M%S")

from ..core import __NOT_HPI_MODULE__



# shared Rss stuff
from datetime import datetime
from typing import NamedTuple, Optional, List, Dict


class Subscription(NamedTuple):
    title: str
    url: str
    id: str # TODO not sure about it...
    # eh, not all of them got reasonable 'created' time
    created_at: Optional[datetime]
    subscribed: bool=True

from typing import Iterable, Tuple, Sequence

# snapshot of subscriptions at time
SubscriptionState = Tuple[datetime, Sequence[Subscription]]


from ..core import warn_if_empty
@warn_if_empty
def compute_subscriptions(*sources: Iterable[SubscriptionState]) -> List[Subscription]:
    """
    Keeps track of everything I ever subscribed to.
    In addition, keeps track of unsubscribed as well (so you'd remember when and why you unsubscribed)
    """
    from itertools import chain
    states = list(chain.from_iterable(sources))
    # TODO keep 'source'/'provider'/'service' attribute?

    by_url: Dict[str, Subscription] = {}
    # ah. dates are used for sorting
    for when, state in sorted(states):
        # TODO use 'when'?
        for feed in state:
            if feed.url not in by_url:
                by_url[feed.url] = feed

    if len(states) == 0:
        return []

    _, last_state = max(states, key=lambda x: x[0])
    last_urls = {f.url for f in last_state}

    res = []
    for u, x in sorted(by_url.items()):
        present = u in last_urls
        res.append(x._replace(subscribed=present))
    return res

from ..core import __NOT_HPI_MODULE__



from my.config import codeforces as config  # type: ignore[attr-defined]


from datetime import datetime, timezone
from functools import cached_property
import json
from typing import NamedTuple, Dict, Iterator


from my.core import get_files, Res
from my.core.konsume import ignore, wrap


Cid = int

class Contest(NamedTuple):
    cid: Cid
    when: datetime

    @classmethod
    def make(cls, j) -> 'Contest':
        return cls(
            cid=j['id'],
            when=datetime.fromtimestamp(j['startTimeSeconds'], tz=timezone.utc),
        )

Cmap = Dict[Cid, Contest]


def get_contests() -> Cmap:
    last = max(get_files(config.export_path, 'allcontests*.json'))
    j = json.loads(last.read_text())
    d = {}
    for c in j['result']:
        cc = Contest.make(c)
        d[cc.cid] = cc
    return d


class Competition(NamedTuple):
    contest_id: Cid
    contest: str
    cmap: Cmap

    @cached_property
    def uid(self) -> Cid:
        return self.contest_id

    def __hash__(self):
        return hash(self.contest_id)

    @cached_property
    def when(self) -> datetime:
        return self.cmap[self.uid].when

    @cached_property
    def summary(self) -> str:
        return f'participated in {self.contest}' # TODO 

    @classmethod
    def make(cls, cmap, json) -> Iterator[Res['Competition']]:
        # TODO try here??
        contest_id = json['contestId'].zoom().value
        contest = json['contestName'].zoom().value
        yield cls(
            contest_id=contest_id,
            contest=contest,
            cmap=cmap,
        )
        # TODO ytry???
        ignore(json, 'rank', 'oldRating', 'newRating')


def data() -> Iterator[Res[Competition]]:
    cmap = get_contests()
    last = max(get_files(config.export_path, 'codeforces*.json'))

    with wrap(json.loads(last.read_text())) as j:
        j['status'].ignore()  # type: ignore[index]
        res = j['result'].zoom()  # type: ignore[index]

        for c in list(res): # TODO maybe we want 'iter' method??
            ignore(c, 'handle', 'ratingUpdateTimeSeconds')
            yield from Competition.make(cmap=cmap, json=c)
            c.consume()
            # TODO maybe if they are all empty, no need to consume??



"""
Monzo transactions data (using https://github.com/karlicoss/monzoexport )
"""
REQUIRES = [
    'git+https://github.com/karlicoss/monzoexport',
]

from dataclasses import dataclass
from pathlib import Path
from typing import Sequence, Iterator

from my.core import (
    Paths,
    get_files,
    make_logger,
)
import my.config


@dataclass
class config(my.config.monzo.monzoexport):
    '''
    Uses [[https://github.com/karlicoss/monzoexport][ghexport]] outputs.
    '''

    export_path: Paths
    '''path[s]/glob to the exported JSON data'''


logger = make_logger(__name__)


def inputs() -> Sequence[Path]:
    return get_files(config.export_path)


import monzoexport.dal as dal


def _dal() -> dal.DAL:
    return dal.DAL(inputs())


def transactions() -> Iterator[dal.MonzoTransaction]:
    return _dal().transactions()



"""
Extracts locations using google_takeout_parser -- no shared code with the deprecated my.location.google
"""

REQUIRES = ["git+https://github.com/seanbreckenridge/google_takeout_parser"]

from typing import Iterator

from my.google.takeout.parser import events, _cachew_depends_on
from google_takeout_parser.models import Location as GoogleLocation

from my.core.common import mcachew, LazyLogger, Stats
from .common import Location

logger = LazyLogger(__name__)


@mcachew(
    depends_on=_cachew_depends_on,
    logger=logger,
)
def locations() -> Iterator[Location]:
    for g in events():
        if isinstance(g, GoogleLocation):
            yield Location(
                lon=g.lng,
                lat=g.lat,
                dt=g.dt,
                accuracy=g.accuracy,
                elevation=None,
                datasource="google_takeout",
            )


def stats() -> Stats:
    from my.core import stat

    return {**stat(locations)}



from typing import Iterable, List
import warnings
from my.core import warn_if_empty


def test_warn_if_empty() -> None:
    @warn_if_empty
    def nonempty() -> Iterable[str]:
        yield 'a'
        yield 'aba'

    @warn_if_empty
    def empty() -> List[int]:
        return []

    # should be rejected by mypy!
    # todo how to actually test it?
    # @warn_if_empty
    # def baad() -> float:
    #     return 0.00

    # reveal_type(nonempty)
    # reveal_type(empty)

    with warnings.catch_warnings(record=True) as w:
        assert list(nonempty()) == ['a', 'aba']
        assert len(w) == 0

        eee = empty()
        assert eee == []
        assert len(w) == 1


def test_warn_iterable() -> None:
    from my.core.common import _warn_iterable
    i1: List[str] = ['a', 'b']
    i2: Iterable[int] = iter([1, 2, 3])
    # reveal_type(i1)
    # reveal_type(i2)
    x1 = _warn_iterable(i1)
    x2 = _warn_iterable(i2)
    # vvvv this should be flagged by mypy
    # _warn_iterable(123)
    # reveal_type(x1)
    # reveal_type(x2)
    with warnings.catch_warnings(record=True) as w:
        assert x1 is i1 # should be unchanged!
        assert len(w) == 0

        assert list(x2) == [1, 2, 3]
        assert len(w) == 0


def test_cachew() -> None:
    from cachew import settings
    settings.ENABLE = True # by default it's off in tests (see conftest.py)

    from my.core.cachew import cache_dir
    from my.core.common import mcachew

    called = 0
    # FIXME ugh. need doublewrap or something
    @mcachew()
    def cf() -> List[int]:
        nonlocal called
        called += 1
        return [1, 2, 3]

    list(cf())
    cc = called
    # todo ugh. how to clean cache?
    # assert called == 1 # precondition, to avoid turdes from previous tests

    assert list(cf()) == [1, 2, 3]
    assert called == cc


def test_cachew_dir_none() -> None:
    from cachew import settings
    settings.ENABLE = True # by default it's off in tests (see conftest.py)

    from my.core.cachew import cache_dir
    from my.core.common import mcachew
    from my.core.core_config import _reset_config as reset
    with reset() as cc:
        cc.cache_dir = None
        called = 0
        @mcachew(cache_path=cache_dir() / 'ctest')
        def cf() -> List[int]:
            nonlocal called
            called += 1
            return [called, called, called]
        assert list(cf()) == [1, 1, 1]
        assert list(cf()) == [2, 2, 2]



"""
Git commits data for repositories on your filesystem
"""
REQUIRES = [
    'gitpython',
]


import shutil
from pathlib import Path
from datetime import datetime, timezone
from dataclasses import dataclass, field
from typing import List, Optional, Iterator, Set, Sequence, cast


from my.core import PathIsh, LazyLogger, make_config
from my.core.cachew import cache_dir
from my.core.common import mcachew
from my.core.warnings import high


from my.config import commits as user_config
@dataclass
class commits_cfg(user_config):
    roots: Sequence[PathIsh] = field(default_factory=list)
    emails: Optional[Sequence[str]] = None
    names: Optional[Sequence[str]] = None


# experiment to make it lazy?
# would be nice to have a nicer syntax for it... maybe make_config could return a 'lazy' object
def config() -> commits_cfg:
    res = make_config(commits_cfg)
    if res.emails is None and res.names is None:
        # todo error policy? throw/warn/ignore
        high("Set either 'emails' or 'names', otherwise you'll get no commits")
    return res

##########################

import git
from git.repo.fun import is_git_dir


log = LazyLogger(__name__, level='info')


def by_me(c: git.objects.commit.Commit) -> bool:
    actor = c.author
    if actor.email in (config().emails or ()):
        return True
    if actor.name in (config().names or ()):
        return True
    return False


@dataclass
class Commit:
    committed_dt: datetime
    authored_dt: datetime
    message: str
    repo: str # TODO put canonical name here straight away??
    sha: str
    ref: Optional[str] = None
    # TODO filter so they are authored by me

    @property
    def dt(self) -> datetime:
        return self.committed_dt

    # for backwards compatibility, was misspelled previously
    @property
    def commited_dt(self) -> datetime:
        high("DEPRECATED! Please replace 'commited_dt' with 'committed_dt' (two 't's instead of one)")
        return self.committed_dt


# TODO not sure, maybe a better idea to move it to timeline?
def fix_datetime(dt: datetime) -> datetime:
    # git module got it's own tzinfo object.. and it's pretty weird
    tz = dt.tzinfo
    assert tz is not None, dt
    assert getattr(tz, '_name') == 'fixed'
    offset = getattr(tz, '_offset')
    ntz = timezone(offset)
    return dt.replace(tzinfo=ntz)


def _git_root(git_dir: PathIsh) -> Path:
    gd = Path(git_dir)
    if gd.name == '.git':
        return gd.parent
    else:
        return gd # must be bare


def _repo_commits_aux(gr: git.Repo, rev: str, emitted: Set[str]) -> Iterator[Commit]:
    # without path might not handle pull heads properly
    for c in gr.iter_commits(rev=rev):
        if not by_me(c):
            continue
        sha = c.hexsha
        if sha in emitted:
            continue
        emitted.add(sha)

        # todo figure out how to handle Union[str, PathLike[Any]].. should it be part of PathIsh?
        repo = str(_git_root(gr.git_dir)) # type: ignore[arg-type]

        yield Commit(
            committed_dt=fix_datetime(c.committed_datetime),
            authored_dt=fix_datetime(c.authored_datetime),
            # hmm no idea why is it typed with Union[str, bytes]??
            # https://github.com/gitpython-developers/GitPython/blob/1746b971387eccfc6fb4e34d3c334079bbb14b2e/git/objects/commit.py#L214
            message=cast(str, c.message).strip(),
            repo=repo,
            sha=sha,
            ref=rev,
        )


def repo_commits(repo: PathIsh):
    gr = git.Repo(str(repo))
    emitted: Set[str] = set()
    for r in gr.references:
        yield from _repo_commits_aux(gr=gr, rev=r.path, emitted=emitted)


def canonical_name(repo: Path) -> str:
    # TODO could determine origin?
    if repo.match('github/repositories/*/repository'):
        return repo.parent.name
    else:
        return repo.name
        # if r.name == 'repository': # 'repository' thing from github..
        #     rname = r.parent.name
        # else:
        #     rname = r.name
    # if 'backups/github' in repo:
    #     pass # TODO 


def _fd_path() -> str:
    # todo move it to core
    fd_path: Optional[str] = shutil.which("fdfind") or shutil.which("fd-find") or shutil.which("fd")
    if fd_path is None:
        high("my.coding.commits requires 'fd' to be installed, See https://github.com/sharkdp/fd#installation")
    assert fd_path is not None
    return fd_path


def git_repos_in(roots: List[Path]) -> List[Path]:
    from subprocess import check_output
    outputs = check_output([
        _fd_path(),
        # '--follow', # right, not so sure about follow... make configurable?
        '--hidden',
        '--no-ignore',  # otherwise doesn't go inside .git directory (from fd v9)
        '--full-path',
        '--type', 'f',
        '/HEAD', # judging by is_git_dir, it should always be here..
        *roots,
    ]).decode('utf8').splitlines()

    candidates = set(Path(o).resolve().absolute().parent for o in outputs)

    # exclude stuff within .git dirs (can happen for submodules?)
    candidates = {c for c in candidates if '.git' not in c.parts[:-1]}

    candidates = {c for c in candidates if is_git_dir(c)}

    repos = list(sorted(map(_git_root, candidates)))
    return repos


def repos() -> List[Path]:
    return git_repos_in(list(map(Path, config().roots)))


# returns modification time for an index to use as hash function
def _repo_depends_on(_repo: Path) -> int:
    for pp in {
        ".git/FETCH_HEAD",
        ".git/HEAD",
        "FETCH_HEAD",  # bare
        "HEAD",  # bare
    }:
        ff = _repo / pp
        if ff.exists():
            return int(ff.stat().st_mtime)
    else:
        raise RuntimeError(f"Could not find a FETCH_HEAD/HEAD file in {_repo}")


def _commits(_repos: List[Path]) -> Iterator[Commit]:
    for r in _repos:
        yield from _cached_commits(r)


def _cached_commits_path(p: Path) -> str:
    p = cache_dir() / 'my.coding.commits:_cached_commits' / str(p.absolute()).strip("/")
    p.mkdir(parents=True, exist_ok=True)
    return str(p)


# per-repo commits, to use cachew
@mcachew(
    depends_on=_repo_depends_on,
    logger=log,
    cache_path=_cached_commits_path,
)
def _cached_commits(repo: Path) -> Iterator[Commit]:
    log.debug('processing %s', repo)
    yield from repo_commits(repo)


def commits() -> Iterator[Commit]:
    return _commits(repos())


# TODO enforce read only? although it doesn't touch index



from __future__ import annotations
from dataclasses import dataclass
from typing import Optional, Callable, Sequence, Iterator, List, Union
from datetime import datetime, timedelta, timezone

from ..common import LocationProtocol, Location
DateExact = Union[datetime, float, int]  # float/int as epoch timestamps

Second = float

@dataclass
class FallbackLocation(LocationProtocol):
    lat: float
    lon: float
    dt: datetime
    duration: Optional[Second] = None
    accuracy: Optional[float] = None
    elevation: Optional[float] = None
    datasource: Optional[str] = None  # which module provided this, useful for debugging

    def to_location(self, end: bool = False) -> Location:
        '''
        by default the start date is used for the location
        If end is True, the start date + duration is used
        '''
        dt: datetime = self.dt
        if end and self.duration is not None:
            dt += timedelta(self.duration)
        return Location(
            lat=self.lat,
            lon=self.lon,
            dt=dt,
            accuracy=self.accuracy,
            elevation=self.elevation,
            datasource=self.datasource,
        )

    @classmethod
    def from_end_date(
        cls,
        *,
        lat: float,
        lon: float,
        dt: datetime,
        end_dt: datetime,
        accuracy: Optional[float] = None,
        elevation: Optional[float] = None,
        datasource: Optional[str] = None,
    ) -> FallbackLocation:
        '''
        Create FallbackLocation from a start date and an end date
        '''
        if end_dt < dt:
            raise ValueError("end_date must be after dt")
        duration = (end_dt - dt).total_seconds()
        return cls(
            lat=lat,
            lon=lon,
            dt=dt,
            duration=duration,
            accuracy=accuracy,
            elevation=elevation,
            datasource=datasource,
        )


# a location estimator can return multiple fallbacks, in case there are
# differing accuracies/to allow for possible matches to be computed
# iteratively
LocationEstimator = Callable[[DateExact], Iterator[FallbackLocation]]
LocationEstimators = Sequence[LocationEstimator]

# helper function, instead of dealing with datetimes while comparing, just use epoch timestamps
def _datetime_timestamp(dt: DateExact) -> float:
    if isinstance(dt, datetime):
        try:
            return dt.timestamp()
        except ValueError:
            # https://github.com/python/cpython/issues/75395
            return dt.replace(tzinfo=timezone.utc).timestamp()
    return float(dt)

def _iter_estimate_from(
    dt: DateExact,
    estimators: LocationEstimators,
) -> Iterator[FallbackLocation]:
    for est in estimators:
        yield from est(dt)


def estimate_from(
    dt: DateExact,
    estimators: LocationEstimators,
    *,
    first_match: bool = False,
    under_accuracy: Optional[int] = None,
) -> Optional[FallbackLocation]:
    '''
    first_match: if True, return the first location found
    under_accuracy: if set, only return locations with accuracy under this value
    '''
    found: List[FallbackLocation] = []
    for loc in _iter_estimate_from(dt, estimators):
        if under_accuracy is not None and loc.accuracy is not None and loc.accuracy > under_accuracy:
            continue
        if first_match:
            return loc
        found.append(loc)

    if not found:
        return None

    # if all items have accuracy, return the one with the lowest accuracy
    # otherwise, we should prefer the order that the estimators are passed in as
    if all(loc.accuracy is not None for loc in found):
        # return the location with the lowest accuracy
        return min(found, key=lambda loc: loc.accuracy)  # type: ignore[return-value, arg-type]
    else:
        # return the first location
        return found[0]



'''
[[https://github.com/jonasoreland/runnerup][Runnerup]] exercise data (TCX format)
'''

REQUIRES = [
    'python-tcxparser',
]

from datetime import timedelta
from pathlib import Path
from typing import Iterable

from .core import Res, get_files
from .core.common import Json
from .core.compat import fromisoformat

import tcxparser  # type: ignore[import-untyped]

from my.config import runnerup as config


# TODO later, use a proper namedtuple?
Workout = Json


def _parse(f: Path) -> Workout:
    tcx = tcxparser.TCXParser(str(f))

    sport = f.stem.split('_')[-1] # todo not sure how reliable...
    hr_avg = tcx.hr_avg

    distance_m = tcx.distance
    duration_s = tcx.duration
    # kmh to match endomondo.. should probably be CI
    speed_avg_kmh = (distance_m / 1000) / (duration_s / 3600)

    # eh. not sure if there is a better way
    # for now use this to be compatible with Endomondo
    # https://beepb00p.xyz/heartbeats_vs_kcals.html
    # filtered for Endomondo running:
    reg_coeff = 0.0993
    intercept = -11.0739
    total_beats = hr_avg * (duration_s / 60)
    kcal_estimate = total_beats * reg_coeff + intercept

    return {
        'id'            : f.name, # not sure?
        'start_time'    : fromisoformat(tcx.started_at),
        'duration'      : timedelta(seconds=tcx.duration),
        'sport'         : sport,
        'heart_rate_avg': tcx.hr_avg,
        'speed_avg'     : speed_avg_kmh,
        'kcal'          : kcal_estimate,
    }
    # from more_itertools import zip_equal
    # for ts, latlon, hr in zip_equal(
    #         tcx.time_values(),
    #         tcx.position_values(),
    #         tcx.hr_values(),
    #         # todo cadence?
    # ):
    #     t = fromisoformat(ts)


def workouts() -> Iterable[Res[Workout]]:
    for f in get_files(config.export_path):
        try:
            yield _parse(f)
        except Exception as e:
            yield e


from .core.pandas import DataFrameT, check_dataframe, error_to_row
@check_dataframe
def dataframe() -> DataFrameT:
    def it():
        for w in workouts():
            if isinstance(w, Exception):
                yield error_to_row(w)
            else:
                yield w
    import pandas as pd
    df = pd.DataFrame(it())
    if 'error' not in df:
        df['error'] = None
    return df


from .core import stat, Stats
def stats() -> Stats:
    return stat(dataframe)



"""
Feedbin RSS reader
"""

from my.config import feedbin as config

from pathlib import Path
from typing import Sequence

from ..core.common import listify, get_files
from ..core.compat import fromisoformat
from .common import Subscription


def inputs() -> Sequence[Path]:
    return get_files(config.export_path)


import json

@listify
def parse_file(f: Path):
    raw = json.loads(f.read_text())
    for r in raw:
        yield Subscription(
            created_at=fromisoformat(r['created_at']),
            title=r['title'],
            url=r['site_url'],
            id=r['id'],
        )


from typing import Iterable
from .common import SubscriptionState
def states() -> Iterable[SubscriptionState]:
    # meh
    from dateutil.parser import isoparse
    for f in inputs():
        # TODO ugh. depends on my naming. not sure if useful?
        dts = f.stem.split('_')[-1]
        dt = isoparse(dts)
        subs = parse_file(f)
        yield dt, subs


def stats():
    from more_itertools import ilen, last
    return {
        'subscriptions': ilen(last(states())[1])
    }



'''
NOTE: you shouldn't modify this file.
You probably want to edit your personal config (check via 'hpi config check' or create with 'hpi config create').

See https://github.com/karlicoss/HPI/blob/master/doc/SETUP.org#setting-up-modules for info on creating your own config

This file is used for:
- documentation (as an example of the config structure)
- mypy: this file provides some type annotations
- for loading the actual user config
'''
#### NOTE: you won't need this line VVVV in your personal config
from my.core import init
###


from datetime import tzinfo
from pathlib import Path
from typing import List


from my.core import Paths, PathIsh


class hypothesis:
    # expects outputs from https://github.com/karlicoss/hypexport
    # (it's just the standard Hypothes.is export format)
    export_path: Paths = r'/path/to/hypothesis/data'

class instapaper:
    export_path: Paths = ''

class smscalls:
    export_path: Paths = ''

class pocket:
    export_path: Paths = ''

class github:
    export_path: Paths = ''

    gdpr_dir: Paths = ''

class reddit:
    class rexport:
        export_path: Paths = ''
    class pushshift:
        export_path: Paths = ''
    class gdpr:
        export_path: Paths = ''

class endomondo:
    export_path: Paths = ''

class exercise:
    workout_log: PathIsh = '/some/path.org'

class bluemaestro:
    export_path: Paths = ''

class stackexchange:
    export_path: Paths = ''

class goodreads:
    export_path: Paths = ''

class pinboard:
    export_dir: Paths = ''

class google:
    class maps:
        class android:
            export_path: Paths = ''

    takeout_path: Paths = ''


from typing import Sequence, Union, Tuple
from datetime import datetime, date, timedelta
DateIsh = Union[datetime, date, str]
LatLon = Tuple[float, float]
class location:
    # todo ugh, need to think about it... mypy wants the type here to be general, otherwise it can't deduce
    # and we can't import the types from the module itself, otherwise would be circular. common module?
    home: Union[LatLon, Sequence[Tuple[DateIsh, LatLon]]] = (1.0, -1.0)
    home_accuracy = 30_000.0

    class via_ip:
        accuracy: float
        for_duration: timedelta

    class gpslogger:
        export_path: Paths = ''
        accuracy: float

    class google_takeout_semantic:
        # a value between 0 and 100, 100 being the most confident
        # set to 0 to include all locations
        # https://locationhistoryformat.com/reference/semantic/#/$defs/placeVisit/properties/locationConfidence
        require_confidence: float = 40
        # default accuracy for semantic locations
        accuracy: float = 100


from typing import Literal
class time:
    class tz:
        policy: Literal['keep', 'convert', 'throw']

        class via_location:
            fast: bool
            sort_locations: bool
            require_accuracy: float


class orgmode:
    paths: Paths


class arbtt:
    logfiles: Paths


from typing import Optional
class commits:
    emails: Optional[Sequence[str]]
    names: Optional[Sequence[str]]
    roots: Sequence[PathIsh]


class pdfs:
    paths: Paths


class zulip:
    class organization:
        export_path: Paths


class bumble:
    class android:
        export_path: Paths


class tinder:
    class android:
        export_path: Paths


class instagram:
    class android:
        export_path: Paths
        username: Optional[str]
        full_name: Optional[str]

    class gdpr:
        export_path: Paths


class hackernews:
    class dogsheep:
        export_path: Paths


class materialistic:
    export_path: Paths


class fbmessenger:
    class fbmessengerexport:
        export_db: PathIsh
        facebook_id: Optional[str]
    class android:
        export_path: Paths


class twitter_archive:
    export_path: Paths


class twitter:
    class talon:
        export_path: Paths
    class android:
        export_path: Paths


class twint:
    export_path: Paths


class browser:
    class export:
        export_path: Paths = ''
    class active_browser:
        export_path: Paths = ''


class telegram:
    class telegram_backup:
        export_path: PathIsh = ''


class demo:
    data_path: Paths
    username: str
    timezone: tzinfo


class simple:
    count: int


class vk_messages_backup:
    storage_path: Path
    user_id: int


class kobo:
    export_path: Paths


class feedly:
    export_path: Paths


class feedbin:
    export_path: Paths


class taplog:
    export_path: Paths


class lastfm:
    export_path: Paths


class rescuetime:
    export_path: Paths


class runnerup:
    export_path: Paths


class emfit:
    export_path: Path
    timezone: tzinfo
    excluded_sids: List[str]


class foursquare:
    export_path: Paths


class rtm:
    export_path: Paths


class imdb:
    export_path: Paths


class roamresearch:
    export_path: Paths
    username: str


class whatsapp:
    class android:
        export_path: Paths
        my_user_id: Optional[str]


class harmonic:
    export_path: Paths


class monzo:
    class monzoexport:
        export_path: Paths



#!/usr/bin/env python3
# see https://github.com/karlicoss/pymplate for up-to-date reference

from setuptools import setup, find_namespace_packages # type: ignore

INSTALL_REQUIRES = [
    'pytz',           # even though it's not needed by the core, it's so common anyway...
    'appdirs',        # very common, and makes it portable
    'more-itertools', # it's just too useful and very common anyway
    'decorator'     , # less pain in writing correct decorators. very mature and stable, so worth keeping in core
    'click>=8.1'    , # for the CLI, printing colors, decorator-based - may allow extensions to CLI
    'kompress'      , # for transparent access to compressed files via pathlib.Path
]


def main() -> None:
    pkg = 'my'
    subpackages = find_namespace_packages('.', include=('my.*',))
    setup(
        name='HPI', # NOTE: 'my' is taken for PyPi already, and makes discovering the project impossible. so we're using HPI
        use_scm_version={
            # todo eh? not sure if I should just rely on proper tag naming and use use_scm_version=True
            # 'version_scheme': 'python-simplified-semver',
            'local_scheme': 'dirty-tag',
        },
        setup_requires=['setuptools_scm'],

        zip_safe=False,

        # eh. find_packages doesn't find anything
        # find_namespace_packages can't find single file packages (like my/common.py)
        packages=[pkg, *subpackages],
        package_data={
            pkg: [
                # for mypy
                'py.typed',
            ],
        },


        url='https://github.com/karlicoss/HPI',
        author='Dmitrii Gerasimov',
        author_email='karlicoss@gmail.com',
        description='A Python interface to my life',

        python_requires='>=3.8',
        install_requires=INSTALL_REQUIRES,
        extras_require={
            'testing': [
                'pytest<8',  # FIXME <8 is temporary workaround till we fix collection with pytest 8; see https://docs.pytest.org/en/stable/changelog.html#collection-changes
                'ruff',
                'mypy',
                'lxml', # for mypy coverage

                # used in some tests.. although shouldn't rely on it
                'pandas',
            ],
            'optional': [
                # todo document these?
                'orjson',  # for my.core.serialize
                'pyfzf_iter',  # for my.core.denylist
                'cachew>=0.15.20231019 ',
                'mypy', # used for config checks
                'colorlog',  # for colored logs
                'enlighten',  # for CLI progress bars
            ],
        },
        entry_points={'console_scripts': ['hpi=my.core.__main__:main']},
    )


if __name__ == '__main__':
    import argparse
    p = argparse.ArgumentParser()
    p.add_argument('--dependencies-only', action='store_true')
    args, _ = p.parse_known_args()
    if args.dependencies_only:
        cmd = ['pip3', 'install', '--user', *INSTALL_REQUIRES]
        scmd = ' '.join(cmd)
        import os
        xx = input(f'Run {scmd} [y/n] ')
        if xx.strip() == 'y':
            os.execvp(
                'pip3',
                cmd
            )
    else:
        main()

# TODO assert??? diff -bur my/ ~/.local/lib/python3.8/site-packages/my/



# Defines some shared config for tests

from datetime import datetime, date, timezone
from pathlib import Path

from typing import Any, NamedTuple
import my.time.tz.via_location as LTZ
from more_itertools import one


class SharedConfig(NamedTuple):
    google: Any
    location: Any
    time: Any


def _prepare_google_config(tmp_path: Path):
    from my.tests.common import testdata
    try:
        track = one(testdata().rglob('italy-slovenia-2017-07-29.json'))
    except ValueError:
        raise RuntimeError('testdata not found, setup git submodules?')


    # todo ugh. unnecessary zipping, but at the moment takeout provider doesn't support plain dirs
    import zipfile
    with zipfile.ZipFile(tmp_path / 'takeout.zip', 'w') as zf:
        zf.writestr('Takeout/Location History/Location History.json', track.read_bytes())

    class google_config:
        takeout_path = tmp_path
    return google_config


# pass tmp_path from pytest to this helper function
# see tests/tz.py as an example
def temp_config(temp_path: Path) -> Any:
    from my.tests.common import reset_modules
    reset_modules()

    LTZ.config.fast = True

    class location:
        home_accuracy = 30_000
        home = (
            # supports ISO strings
            ('2005-12-04'                                       , (42.697842, 23.325973)), # Bulgaria, Sofia
            # supports date/datetime objects
            (date(year=1980, month=2, day=15)                   , (40.7128  , -74.0060 )), # NY
            # check tz handling..
            (datetime.fromtimestamp(1600000000, tz=timezone.utc), (55.7558  , 37.6173  )), # Moscow, Russia
        )
        # note: order doesn't matter, will be sorted in the data provider
        class via_ip:
            accuracy = 15_000
        class gpslogger:
            pass

    class time:
        class tz:
            class via_location:
                pass # just rely on the defaults...


    return SharedConfig(google=_prepare_google_config(temp_path), location=location, time=time)



"""
Bumble data from Android app database (in =/data/data/com.instagram.android/databases/direct.db=)
"""
from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
import json
from pathlib import Path
import sqlite3
from typing import Iterator, Sequence, Optional, Dict, Union

from my.core import (
    get_files,
    Paths,
    make_config,
    make_logger,
    datetime_naive,
    Json,
    Res,
    assert_never,
)
from my.core.common import unique_everseen
from my.core.cachew import mcachew
from my.core.error import echain
from my.core.sqlite import sqlite_connect_immutable, select

from my.config import instagram as user_config


logger = make_logger(__name__)


@dataclass
class instagram_android_config(user_config.android):
    # paths[s]/glob to the exported sqlite databases
    export_path: Paths

    # sadly doesn't seem easy to extract user's own handle/name from the db...
    # todo maybe makes more sense to keep in parent class? not sure...
    username: Optional[str] = None
    full_name: Optional[str] = None


config = make_config(instagram_android_config)


def inputs() -> Sequence[Path]:
    return get_files(config.export_path)


@dataclass(unsafe_hash=True)
class User:
    id: str
    username: str
    full_name: str


# todo not sure about order of fields...
@dataclass
class _BaseMessage:
    id: str
    # NOTE: ffs, looks like they keep naive timestamps in the db (checked some random messages)
    created: datetime_naive
    text: str
    thread_id: str


@dataclass(unsafe_hash=True)
class _Message(_BaseMessage):
    user_id: str
    # TODO ugh. can't figure out if dms have proper replies?
    # reply_to_id: Optional[str]


@dataclass(unsafe_hash=True)
class Message(_BaseMessage):
    user: User
    # TODO could also extract Thread object? not sure if useful
    # reply_to: Optional[Message]


# this is kinda expecrimental
# basically just using RuntimeError(msg_id, *rest) has an unfortunate consequence:
# there are way too many 'similar' errors (on different msg_id)
# however passing msg_id is nice as a means of supplying extra context
# so this is a compromise, the 'duplicate' errors will be filtered out by unique_everseen


class MessageError(RuntimeError):
    def __init__(self, msg_id: str, *rest: str) -> None:
        super().__init__(msg_id, *rest)
        self.rest = rest

    def __hash__(self, other):
        return hash(self.rest)

    def __eq__(self, other) -> bool:
        if not isinstance(other, MessageError):
            return False
        return self.rest == other.rest


def _parse_message(j: Json) -> Optional[_Message]:
    id = j['item_id']
    t = j['item_type']
    tid = j['thread_key']['thread_id']
    uid = j['user_id']
    created = datetime.fromtimestamp(int(j['timestamp']) / 1_000_000)
    text: Optional[str] = None
    if t == 'text':
        text = j['text']
    elif t == 'reel_share':
        # TODO include reel_share -> media??
        # the problem is that the links are deliberately expired by instagram..
        text = j['reel_share']['text']
    elif t == 'action_log':
        # for likes this ends up as 'Liked a message' or reactions
        # which isn't super useful by itself perhaps, but matches GDPR so lets us unify threads better
        text = j['action_log']['description']
    else:
        raise MessageError(id, f"{t} isn't handled yet")

    assert text is not None, j

    return _Message(
        id=id,
        created=created,
        text=text,
        thread_id=tid,
        user_id=uid,
        # reply_to_id='FIXME',
    )


def _process_db(db: sqlite3.Connection) -> Iterator[Res[Union[User, _Message]]]:
    # TODO ugh. seems like no way to extract username?
    # sometimes messages (e.g. media_share) contain it in message field
    # but generally it's not present. ugh
    for (self_uid,) in select(('user_id',), 'FROM session', db=db):
        yield User(
            id=str(self_uid),
            full_name=config.full_name or 'USERS_OWN_FULL_NAME',
            username=config.full_name or 'USERS_OWN_USERNAME',
        )

    for (thread_json,) in select(('thread_info',), 'FROM threads', db=db):
        j = json.loads(thread_json)
        # todo in principle should leave the thread attached to the message?
        # since thread is a group of users?
        pre_users = []
        # inviter usually contains our own user
        if 'inviter' in j:
            # sometimes it's missing (e.g. in broadcast channels)
            pre_users.append(j['inviter'])
        pre_users.extend(j['recipients'])
        for r in pre_users:
            # id disappeared and seems that pk_id is in use now (around december 2022)
            uid = r.get('id') or r.get('pk_id')
            assert uid is not None
            yield User(
                id=str(uid),  # for some reason it's int in the db
                full_name=r['full_name'],
                username=r['username'],
            )

    for (msg_json,) in select(('message',), 'FROM messages ORDER BY timestamp', db=db):
        # eh, seems to contain everything in json?
        j = json.loads(msg_json)
        try:
            m = _parse_message(j)
            if m is not None:
                yield m
        except Exception as e:
            yield e


def _entities() -> Iterator[Res[Union[User, _Message]]]:
    # NOTE: definitely need to merge multiple, app seems to recycle old messages
    # TODO: hmm hard to guarantee timestamp ordering when we use synthetic input data...
    # todo use TypedDict?
    paths = inputs()
    total = len(paths)
    width = len(str(total))
    for idx, path in enumerate(paths):
        logger.info(f'processing [{idx:>{width}}/{total:>{width}}] {path}')
        with sqlite_connect_immutable(path) as db:
            try:
                yield from _process_db(db=db)
            except Exception as e:
                # todo use error policy here
                yield echain(RuntimeError(f'While processing {path}'), cause=e)


@mcachew(depends_on=inputs)
def messages() -> Iterator[Res[Message]]:
    id2user: Dict[str, User] = {}
    for x in unique_everseen(_entities):
        if isinstance(x, Exception):
            yield x
            continue
        if isinstance(x, User):
            id2user[x.id] = x
            continue
        if isinstance(x, _Message):
            try:
                user = id2user[x.user_id]
            except Exception as e:
                yield e
                continue
            yield Message(
                id=x.id,
                created=x.created,
                text=x.text,
                thread_id=x.thread_id,
                user=user,
            )
            continue
        assert_never(x)



'''
Just a demo module for testing and documentation purposes
'''

from .core import Paths, PathIsh

from typing import Optional
from datetime import tzinfo, timezone

from my.config import demo as user_config
from dataclasses import dataclass


@dataclass
class demo(user_config):
    data_path: Paths
    username: str
    timezone: tzinfo = timezone.utc

    external: Optional[PathIsh] = None

    @property
    def external_module(self):
        rpath = self.external
        if rpath is not None:
            from .core.common import import_dir
            return import_dir(rpath)

        import my.config.repos.external as m # type: ignore
        return m


from .core import make_config
config = make_config(demo)

# TODO not sure about type checking?
external = config.external_module


from pathlib import Path
from typing import Sequence, Iterable
from datetime import datetime
from .core import Json, get_files

@dataclass
class Item:
    '''
    Some completely arbitrary artificial stuff, just for testing
    '''
    username: str
    raw: Json
    dt: datetime


def inputs() -> Sequence[Path]:
    return get_files(config.data_path)


import json
def items() -> Iterable[Item]:
    for f in inputs():
        dt = datetime.fromtimestamp(f.stat().st_mtime, tz=config.timezone)
        j = json.loads(f.read_text())
        for raw in j:
            yield Item(
                username=config.username,
                raw=external.identity(raw),
                dt=dt,
            )



from pathlib import Path

import pytest

from my.calendar.holidays import is_holiday


def test() -> None:
    assert is_holiday('20190101')
    assert not is_holiday('20180601')
    assert is_holiday('20200906') # national holiday in Bulgaria


@pytest.fixture(autouse=True)
def prepare(tmp_path: Path):
    from . import tz
    # todo meh. fixtures can't be called directly?
    orig = tz.prepare.__wrapped__ # type: ignore
    yield from orig(tmp_path)



"""
Facebook Messenger messages

Uses the output of [[https://github.com/karlicoss/fbmessengerexport][fbmessengerexport]]
"""
REQUIRES = [
    'git+https://github.com/karlicoss/fbmessengerexport',
]

from contextlib import ExitStack, contextmanager
from dataclasses import dataclass
from typing import Iterator

from my.core import PathIsh, Res, stat, Stats
from my.core.warnings import high
from my.config import fbmessenger as user_config

import fbmessengerexport.dal as messenger


###
# support old style config
_new_section = getattr(user_config, 'fbmessengerexport', None)
_old_attr    = getattr(user_config, 'export_db', None)

if _new_section is None and _old_attr is not None:
    high("""DEPRECATED! Please modify your fbmessenger config to look like:

class fbmessenger:
    class fbmessengerexport:
        export_db: PathIsh = '/path/to/fbmessengerexport/database'
            """)
    class fbmessengerexport:
        export_db = _old_attr
    setattr(user_config, 'fbmessengerexport', fbmessengerexport)
###


@dataclass
class config(user_config.fbmessengerexport):
    export_db: PathIsh


@contextmanager
def _dal() -> Iterator[messenger.DAL]:
    model = messenger.DAL(config.export_db)
    with ExitStack() as stack:
        if hasattr(model, '__dal__'):  # defensive to support legacy fbmessengerexport
            stack.enter_context(model)
        yield model


def messages() -> Iterator[Res[messenger.Message]]:
    with _dal() as model:
        for t in model.iter_threads():
            yield from t.iter_messages()


def stats() -> Stats:
    return stat(messages)



print(f'[main] {__name__} hello')

from typing import Protocol

class Source(Protocol):
    def tweets(self) -> list[str]:
        ...

def merge(*sources: Source) -> list[str]:
    from itertools import chain
    return list(chain.from_iterable(src.tweets() for src in sources))



"""
[[https://rememberthemilk.com][Remember The Milk]] tasks and notes
"""

REQUIRES = [
    'icalendar',
]

from datetime import datetime
from functools import cached_property
import re
from typing import Dict, List, Iterator

from .core.common import LazyLogger, get_files, group_by_key, make_dict

from my.config import rtm as config


import icalendar # type: ignore
from icalendar.cal import Todo # type: ignore


logger = LazyLogger(__name__)


# TODO extract in a module to parse RTM's ical?
class MyTodo:
    def __init__(self, todo: Todo, revision=None) -> None:
        self.todo = todo
        self.revision = revision

    @cached_property
    def notes(self) -> List[str]:
        # TODO can there be multiple??
        desc = self.todo['DESCRIPTION']
        notes = re.findall(r'---\n\n(.*?)\n\nUpdated:', desc, flags=re.DOTALL)
        return notes

    @cached_property
    def tags(self) -> List[str]:
        desc = self.todo['DESCRIPTION']
        [tags_str] = re.findall(r'\nTags: (.*?)\n', desc, flags=re.DOTALL)
        if tags_str == 'none':
            return []
        tags = [t.strip() for t in tags_str.split(',')]
        return tags

    @cached_property
    def uid(self) -> str:
        return str(self.todo['UID'])

    @cached_property
    def title(self) -> str:
        return str(self.todo['SUMMARY'])

    def get_status(self) -> str:
        if 'STATUS' not in self.todo:
            return None # type: ignore
        # TODO 'COMPLETED'? 
        return str(self.todo['STATUS'])

    # TODO tz?
    @cached_property
    def time(self) -> datetime:
        t1 = self.todo['DTSTAMP'].dt
        t2 = self.todo['LAST-MODIFIED'].dt
        assert t1 == t2 # TODO not sure which one is correct
        return t1

    def is_completed(self) -> bool:
        return self.get_status() == 'COMPLETED'

    def __repr__(self):
        return repr(self.todo)

    def __str__(self):
        return str(self.todo)

    @staticmethod
    def alala_key(mtodo):
        return (mtodo.revision, mtodo.get_time())


class DAL:
    def __init__(self, data: str, revision=None) -> None:
        self.cal = icalendar.Calendar.from_ical(data)
        self.revision = revision

    def all_todos(self) -> Iterator[MyTodo]:
        for t in self.cal.walk('VTODO'):
            yield MyTodo(t, self.revision)

    def get_todos_by_uid(self) -> Dict[str, MyTodo]:
        todos = self.all_todos()
        return make_dict(todos, key=lambda t: t.uid)

    def get_todos_by_title(self) -> Dict[str, List[MyTodo]]:
        todos = self.all_todos()
        return group_by_key(todos, lambda todo: todo.title)


def dal():
    last = get_files(config.export_path)[-1]
    data = last.read_text()
    return DAL(data=data, revision='TODO')


def all_tasks() -> Iterator[MyTodo]:
    yield from dal().all_todos()


def active_tasks() -> Iterator[MyTodo]:
    for t in all_tasks():
        if not t.is_completed():
            yield t




from __future__ import annotations

from typing import Dict, Any, List, Iterable
import json
from functools import lru_cache
from datetime import datetime, date, time, timedelta
from pathlib import Path

import pytz

from ..core.common import LazyLogger

logger = LazyLogger(__name__)

from my.config import jawbone as config  # type: ignore[attr-defined]


BDIR = config.export_dir
PHASES_FILE = BDIR / 'phases.json'
SLEEPS_FILE = BDIR / 'sleeps.json'
GRAPHS_DIR = BDIR / 'graphs'



XID = str # TODO how to shared with backup thing?

Phases = Dict[XID, Any]
@lru_cache(1)
def get_phases() -> Phases:
    return json.loads(PHASES_FILE.read_text())

# TODO use awakenings and quality
class SleepEntry:
    def __init__(self, js) -> None:
        self.js = js

    @property
    def date_(self) -> date:
        return self.sleep_end.date()

    def _fromts(self, ts: int) -> datetime:
        return datetime.fromtimestamp(ts, tz=self._tz)

    @property
    def _tz(self):
        return pytz.timezone(self._details['tz'])

    @property
    def title(self) -> str:
        return self.js['title']

    @property
    def xid(self) -> XID:
        return self.js['xid']

    @property
    def _details(self):
        return self.js['details']

    # TODO figure out timezones..
    # not sure how.. I guess by the american ones
    @property
    def created(self) -> datetime:
        return self._fromts(self.js['time_created'])

    @property
    def completed(self) -> datetime:
        return self._fromts(self.js['time_completed'])

    @property
    def asleep(self) -> datetime:
        return self._fromts(self._details['asleep_time'])

    @property
    def sleep_start(self) -> datetime:
        return self.asleep # TODO careful, maybe use same logic as emfit

    @property
    def bed_time(self) -> int:
        return int((self.sleep_end - self.sleep_start).total_seconds()) // 60

    @property
    def sleep_end(self) -> datetime:
        return self._fromts(self._details['awake_time'])

    @property
    def graph(self) -> Path:
        return GRAPHS_DIR / (self.xid + ".png")

    # TODO might be useful to cache these??
    @property
    def phases(self) -> List[datetime]:
        # TODO make sure they are consistent with emfit?
        return [self._fromts(i['time']) for i in get_phases()[self.xid]]

    def __str__(self) -> str:
        return f"{self.date_.strftime('%a %d %b')} {self.title}"

    def __repr__(self) -> str:
        return str(self)


def load_sleeps() -> List[SleepEntry]:
    sleeps = json.loads(SLEEPS_FILE.read_text())
    return [SleepEntry(js) for js in sleeps]


from ..core.error import Res, set_error_datetime, extract_error_datetime

def pre_dataframe() -> Iterable[Res[SleepEntry]]:
    sleeps = load_sleeps()
    # todo emit error if graph doesn't exist??
    sleeps = [s for s in sleeps if s.graph.exists()] # TODO careful..
    from ..core.common import group_by_key
    for dd, group in group_by_key(sleeps, key=lambda s: s.date_).items():
        if len(group) == 1:
            yield group[0]
        else:
            err = RuntimeError(f'Multiple sleeps per night, not supported yet: {group}')
            set_error_datetime(err, dt=dd)  # type: ignore[arg-type]
            logger.exception(err)
            yield err


def dataframe():
    dicts: List[Dict[str, Any]] = []
    for s in pre_dataframe():
        d: Dict[str, Any]
        if isinstance(s, Exception):
            dt = extract_error_datetime(s)
            d = {
                'date' : dt,
                'error': str(s),
            }
        else:
            d = {
                # TODO make sure sleep start/end are consistent with emfit? add to test as well..
                # I think it makes sense to be end date as 99% of time
                # or maybe I shouldn't care about this at all?
                'date'       : s.date_,
                'sleep_start': s.sleep_start,
                'sleep_end'  : s.sleep_end,
                'bed_time'   : s.bed_time,
            }
        dicts.append(d)

    import pandas as pd
    return pd.DataFrame(dicts)
    # TODO tz is in sleeps json


def stats():
    from ..core import stat
    return stat(pre_dataframe)


#### NOTE: most of the stuff below is deprecated and remnants of my old code!
#### sorry for it, feel free to remove if you don't need it


def hhmm(time: datetime):
    return time.strftime("%H:%M")


# def xpos(time: datetime) -> float:
#     tick = span / width
#     fromstart = time - sleep.created
#     return fromstart / tick


def plot_one(sleep: SleepEntry, fig, axes, xlims=None, showtext=True):
    import matplotlib.dates as mdates  # type: ignore[import-not-found]

    span = sleep.completed - sleep.created
    print(f"{sleep.xid} span: {span}")

    # pip install imageio
    from imageio import imread # type: ignore

    img = imread(sleep.graph)
    # all of them are 300x300 images apparently
    # span for image
    xspan = [sleep.created, sleep.completed]
    xspan = [mdates.date2num(i) for i in xspan]
    if xlims is None:
        tt = sleep.created
        hour = tt.hour
        # TODO maybe assert that hour is somewhere between 20 and 8 or something
        start: datetime
        starttime = time(23, 00)
        if hour >= 20:
            # went to bed before midnight
            start = datetime.combine(tt.date(), starttime)
        elif hour <= 8:
            # went to bed after midnight
            start = datetime.combine(tt.date() - timedelta(days=1), starttime)
        else:
            print("wtf??? weird time for sleep...")
            # choosing at random
            start = datetime.combine(tt.date(), starttime)
        end = start + timedelta(hours=10)
        xlims = [start, end]

    # axes.figure(figsize=(10, 5))
    axes.set_xlim(xlims)
    hhmm_fmt = mdates.DateFormatter('%H:%M')
    axes.xaxis.set_major_formatter(hhmm_fmt)
    ticks = sleep.phases if showtext else []
    axes.xaxis.set_ticks(ticks)
    axes.yaxis.set_ticks([])
    axes.tick_params(
        axis='both',
        which='major',
        length=0,
        labelsize=7,
        rotation=30,
        pad=-14, # err... hacky
    )

    ylims = [0, 50]
    axes.set_ylim(ylims)

    axes.imshow(
        img,
        zorder=0,
        extent=[
            xspan[0], xspan[1],
            ylims[0], ylims[1],
        ],
        aspect='auto',
    )
    # axes.set_title(str(sleep))
    # axes.title.set_size(10)

    if showtext:
        axes.text(xlims[1] - timedelta(hours=1.5), 20, str(sleep),)
    # plt.text(sleep.asleep(), 0, hhmm(sleep.asleep()))


# TODO not really sure it belongs here...
# import melatonin
# dt = melatonin.get_data()

def predicate(sleep: SleepEntry):
    """
       Filter for comparing similar sleep sessions
    """
    start = sleep.created.time()
    end = sleep.completed.time()
    if (time(23, 0) <= start <= time(23, 30)) and (time(5, 30) <= end <= time(6, 30)):
        return True
    return False


# TODO move to dashboard
def plot() -> None:
    from matplotlib.figure import Figure  # type: ignore[import-not-found]
    import matplotlib.pyplot as plt  # type: ignore[import-not-found]

    # TODO FIXME melatonin data
    melatonin_data = {} # type: ignore[var-annotated]

    # TODO ??
    sleeps = list(filter(predicate, load_sleeps()))
    sleeps_count = len(sleeps)
    print(sleeps_count)

    fig: Figure = plt.figure(figsize=(15, sleeps_count * 1))

    axarr = fig.subplots(nrows=len(sleeps))
    for i, (sleep, axes) in enumerate(zip(sleeps, axarr)):
        plot_one(sleep, fig, axes, showtext=True)
        used = melatonin_data.get(sleep.date_, None)
        sused: str
        color: str
        # used = True if used is None else False # TODO?
        if used is True:
            sused = "YES"
            color = 'green'
        elif used is False:
            sused = "NO"
            color = 'red'
        else:
            sused = "??"
            color = 'white'
        axes.text(axes.get_xlim()[0], 20, sused)
        axes.patch.set_alpha(0.5)
        axes.set_facecolor(color)


    plt.tight_layout()
    plt.subplots_adjust(hspace=0.0)
    # er... this saves with a different aspect ratio for some reason.
    # tap 'ctrl-s' on mpl plot window to save..
    # plt.savefig('res.png', asp)
    plt.show()




'''
The idea of this module is to avoid imports of external HPI modules and code evaluation via ast module etc.

This potentially allows it to be:

- robust: can discover modules that can't be imported, generally makes it foolproof
- faster: importing is slow and with tens of modules can be noteiceable
- secure: can be executed in a sandbox & used during setup

It should be free of external modules, importlib, exec, etc. etc.
'''

REQUIRES = 'REQUIRES'
NOT_HPI_MODULE_VAR = '__NOT_HPI_MODULE__'

###

import ast
import os
from typing import Optional, Sequence, List, NamedTuple, Iterable, cast, Any
from pathlib import Path
import re
import logging

'''
None means that requirements weren't defined (different from empty requirements)
'''
Requires = Optional[Sequence[str]]


class HPIModule(NamedTuple):
    name: str
    skip_reason: Optional[str]
    doc: Optional[str] = None
    file: Optional[Path] = None
    requires: Requires = None
    legacy: Optional[str] = None  # contains reason/deprecation warning


def ignored(m: str) -> bool:
    excluded = [
        # legacy stuff left for backwards compatibility
        'core.*',
        'config.*',
    ]
    exs = '|'.join(excluded)
    return re.match(f'^my.({exs})$', m) is not None


def has_stats(src: Path) -> bool:
    # todo make sure consistent with get_stats?
    return _has_stats(src.read_text())


def _has_stats(code: str) -> bool:
    a: ast.Module = ast.parse(code)
    for x in a.body:
        try: # maybe assign
            [tg] = cast(Any, x).targets
            if tg.id == 'stats':
                return True
        except:
            pass
        try: # maybe def?
            name = cast(Any, x).name
            if name == 'stats':
                return True
        except:
            pass
    return False


def _is_not_module_src(src: Path) -> bool:
    a: ast.Module = ast.parse(src.read_text())
    return _is_not_module_ast(a)


def _is_not_module_ast(a: ast.Module) -> bool:
    marker = NOT_HPI_MODULE_VAR
    return any(
        getattr(node, 'name', None) == marker  # direct definition
        or any(getattr(n, 'name', None) == marker for n in getattr(node, 'names', []))  # import from
        for node in a.body
    )


def _is_legacy_module(a: ast.Module) -> bool:
    marker = 'handle_legacy_import'
    return any(
        getattr(node, 'name', None) == marker  # direct definition
        or any(getattr(n, 'name', None) == marker for n in getattr(node, 'names', []))  # import from
        for node in a.body
    )


# todo should be defensive? not sure
def _extract_requirements(a: ast.Module) -> Requires:
    # find the assignment..
    for x in a.body:
        if not isinstance(x, ast.Assign):
            continue
        tg = x.targets
        if len(tg) != 1:
            continue
        t = tg[0]
        # could be Subscript.. so best to keep dynamic
        id_ = getattr(t, 'id', None)
        if id_ != REQUIRES:
            continue
        vals = x.value
        # could be List/Tuple/Set?
        elts = getattr(vals, 'elts', None)
        if elts is None:
            continue
        deps = []
        for c in elts:
            if isinstance(c, ast.Constant):
                deps.append(c.value)
            elif isinstance(c, ast.Str):
                deps.append(c.s)
            else:
                raise RuntimeError(f"Expecting string constants only in {REQUIRES} declaration")
        return tuple(deps)
    return None


# todo should probably be more defensive..
def all_modules() -> Iterable[HPIModule]:
    """
    Return all importable modules under all items in the 'my' namespace package

    Note: This returns all modules under all roots - if you have
    several overlays (multiple items in my.__path__ and you've overridden
    modules), this can return multiple HPIModule objects with the same
    name. It should respect import order, as we're traversing
    in my.__path__ order, so module_by_name should still work
    and return the correctly resolved module, but all_modules
    can have duplicates
    """
    for my_root in _iter_my_roots():
        yield from _modules_under_root(my_root)


def _iter_my_roots() -> Iterable[Path]:
    import my  # doesn't import any code, because of namespace package

    paths: List[str] = list(my.__path__)
    if len(paths) == 0:
        # should probably never happen?, if this code is running, it was imported
        # because something was added to __path__ to match this name
        raise RuntimeError("my.__path__ was empty, try re-installing HPI?")
    else:
        yield from map(Path, paths)


def _modules_under_root(my_root: Path) -> Iterable[HPIModule]:
    """
    Experimental version, which isn't importing the modules, making it more robust and safe.
    """
    for f in sorted(my_root.rglob('*.py')):
        if f.is_symlink():
            continue  # meh
        mp = f.relative_to(my_root.parent)
        if mp.name == '__init__.py':
            mp = mp.parent
        m = str(mp.with_suffix('')).replace(os.sep, '.')
        if ignored(m):
            continue
        a: ast.Module = ast.parse(f.read_text())

        # legacy modules are 'forced' to be modules so 'hpi module install' still works for older modules
        # a bit messy, will think how to fix it properly later
        legacy_module = _is_legacy_module(a)
        if _is_not_module_ast(a) and not legacy_module:
            continue
        doc = ast.get_docstring(a, clean=False)

        requires: Requires = None
        try:
            requires = _extract_requirements(a)
        except Exception as e:
            logging.exception(e)

        legacy = f'{m} is DEPRECATED. Please refer to the module documentation.' if legacy_module else None

        yield HPIModule(
            name=m,
            skip_reason=None,
            doc=doc,
            file=f.relative_to(my_root.parent),
            requires=requires,
            legacy=legacy,
        )


def module_by_name(name: str) -> HPIModule:
    for m in all_modules():
        if m.name == name:
            return m
    raise RuntimeError(f'No such module: {name}')


### tests


def test() -> None:
    # TODO this should be a 'sanity check' or something
    assert len(list(all_modules())) > 10  # kinda arbitrary


def test_demo() -> None:
    demo = module_by_name('my.demo')
    assert demo.doc is not None
    assert demo.file == Path('my', 'demo.py')
    assert demo.requires is None


def test_excluded() -> None:
    for m in all_modules():
        assert 'my.core.' not in m.name


def test_requires() -> None:
    photos = module_by_name('my.photos.main')
    r = photos.requires
    assert r is not None
    assert len(r) == 2  # fragile, but ok for now


def test_legacy_modules() -> None:
    # shouldn't crash
    module_by_name('my.reddit')
    module_by_name('my.fbmessenger')


def test_pure() -> None:
    """
    We want to keep this module clean of other HPI imports
    """
    # this uses string concatenation here to prevent
    # these tests from testing against themselves
    src = Path(__file__).read_text()
    # 'import my' is allowed, but
    # dont allow anything other HPI modules
    assert re.findall('import ' + r'my\.\S+', src, re.M) == []
    assert 'from ' + 'my' not in src


def test_has_stats() -> None:
    assert not _has_stats('')
    assert not _has_stats('x = lambda : whatever')

    assert _has_stats('''
def stats():
    pass
''')

    assert _has_stats('''
stats = lambda: "something"
''')

    assert _has_stats('''
stats = other_function
    ''')



from glob import glob as do_glob
from pathlib import Path
from datetime import datetime
import functools
from contextlib import contextmanager
import os
import sys
import types
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    NoReturn,
    Optional,
    Sequence,
    TYPE_CHECKING,
    Tuple,
    TypeVar,
    Union,
    cast,
    get_args,
    get_type_hints,
    get_origin,
)
import warnings
from . import warnings as core_warnings

# some helper functions
PathIsh = Union[Path, str]

# TODO only used in tests? not sure if useful at all.
def import_file(p: PathIsh, name: Optional[str] = None) -> types.ModuleType:
    p = Path(p)
    if name is None:
        name = p.stem
    import importlib.util
    spec = importlib.util.spec_from_file_location(name, p)
    assert spec is not None, f"Fatal error; Could not create module spec from {name} {p}"
    foo = importlib.util.module_from_spec(spec)
    loader = spec.loader; assert loader is not None
    loader.exec_module(foo)
    return foo


def import_from(path: PathIsh, name: str) -> types.ModuleType:
    path = str(path)
    try:
        sys.path.append(path)
        import importlib
        return importlib.import_module(name)
    finally:
        sys.path.remove(path)


def import_dir(path: PathIsh, extra: str='') -> types.ModuleType:
    p = Path(path)
    if p.parts[0] == '~':
        p = p.expanduser() # TODO eh. not sure about this..
    return import_from(p.parent, p.name + extra)


T = TypeVar('T')
K = TypeVar('K')
V = TypeVar('V')

# TODO deprecate? more_itertools.one should be used
def the(l: Iterable[T]) -> T:
    it = iter(l)
    try:
        first = next(it)
    except StopIteration:
        raise RuntimeError('Empty iterator?')
    assert all(e == first for e in it)
    return first


# TODO more_itertools.bucket?
def group_by_key(l: Iterable[T], key: Callable[[T], K]) -> Dict[K, List[T]]:
    res: Dict[K, List[T]] = {}
    for i in l:
        kk = key(i)
        lst = res.get(kk, [])
        lst.append(i)
        res[kk] = lst
    return res


def _identity(v: T) -> V:  # type: ignore[type-var]
    return cast(V, v)


# ugh. nothing in more_itertools?
def ensure_unique(
        it: Iterable[T],
        *,
        key: Callable[[T], K],
        value: Callable[[T], V]=_identity,
        key2value: Optional[Dict[K, V]]=None
) -> Iterable[T]:
    if key2value is None:
        key2value = {}
    for i in it:
        k = key(i)
        v = value(i)
        pv = key2value.get(k, None)
        if pv is not None:
            raise RuntimeError(f"Duplicate key: {k}. Previous value: {pv}, new value: {v}")
        key2value[k] = v
        yield i


def test_ensure_unique() -> None:
    import pytest
    assert list(ensure_unique([1, 2, 3], key=lambda i: i)) == [1, 2, 3]

    dups = [1, 2, 1, 4]
    # this works because it's lazy
    it = ensure_unique(dups, key=lambda i: i)

    # but forcing throws
    with pytest.raises(RuntimeError, match='Duplicate key'):
        list(it)

    # hacky way to force distinct objects?
    list(ensure_unique(dups, key=lambda i: object()))


def make_dict(
        it: Iterable[T],
        *,
        key: Callable[[T], K],
        value: Callable[[T], V]=_identity
) -> Dict[K, V]:
    res: Dict[K, V] = {}
    uniques = ensure_unique(it, key=key, value=value, key2value=res)
    for _ in uniques:
        pass  # force the iterator
    return res


def test_make_dict() -> None:
    it = range(5)
    d = make_dict(it, key=lambda i: i, value=lambda i: i % 2)
    assert d == {0: 0, 1: 1, 2: 0, 3: 1, 4: 0}

    # check type inference
    d2: Dict[str, int ] = make_dict(it, key=lambda i: str(i))
    d3: Dict[str, bool] = make_dict(it, key=lambda i: str(i), value=lambda i: i % 2 == 0)


# https://stackoverflow.com/a/12377059/706389
def listify(fn=None, wrapper=list):
    """
    Wraps a function's return value in wrapper (e.g. list)
    Useful when an algorithm can be expressed more cleanly as a generator
    """
    def listify_return(fn):
        @functools.wraps(fn)
        def listify_helper(*args, **kw):
            return wrapper(fn(*args, **kw))
        return listify_helper
    if fn is None:
        return listify_return
    return listify_return(fn)


# todo use in bluemaestro
# def dictify(fn=None, key=None, value=None):
#     def md(it):
#         return make_dict(it, key=key, value=value)
#     return listify(fn=fn, wrapper=md)


from .logging import setup_logger, LazyLogger


Paths = Union[Sequence[PathIsh], PathIsh]


DEFAULT_GLOB = '*'
def get_files(
        pp: Paths,
        glob: str=DEFAULT_GLOB,
        sort: bool=True,
        guess_compression: bool=True,
) -> Tuple[Path, ...]:
    """
    Helper function to avoid boilerplate.

    Tuple as return type is a bit friendlier for hashing/caching, so hopefully makes sense
    """
    # TODO FIXME mm, some wrapper to assert iterator isn't empty?
    sources: List[Path]
    if isinstance(pp, Path):
        sources = [pp]
    elif isinstance(pp, str):
        if pp == '':
            # special case -- makes sense for optional data sources, etc
            return ()  # early return to prevent warnings etc
        sources = [Path(pp)]
    else:
        sources = [p if isinstance(p, Path) else Path(p) for p in pp]

    def caller() -> str:
        import traceback

        # TODO ugh. very flaky... -3 because [<this function>, get_files(), <actual caller>]
        return traceback.extract_stack()[-3].filename

    paths: List[Path] = []
    for src in sources:
        if src.parts[0] == '~':
            src = src.expanduser()
        # note: glob handled first, because e.g. on Windows asterisk makes is_dir unhappy
        gs = str(src)
        if '*' in gs:
            if glob != DEFAULT_GLOB:
                warnings.warn(f"{caller()}: treating {gs} as glob path. Explicit glob={glob} argument is ignored!")
            paths.extend(map(Path, do_glob(gs)))
        elif os.path.isdir(str(src)):
            # NOTE: we're using os.path here on purpose instead of src.is_dir
            # the reason is is_dir for archives might return True and then
            # this clause would try globbing insize the archives
            # this is generally undesirable (since modules handle archives themselves)

            # todo not sure if should be recursive?
            # note: glob='**/*.ext' works without any changes.. so perhaps it's ok as it is
            gp: Iterable[Path] = src.glob(glob)
            paths.extend(gp)
        else:
            assert src.exists(), src
            # todo assert matches glob??
            paths.append(src)

    if sort:
        paths = list(sorted(paths))

    if len(paths) == 0:
        # todo make it conditionally defensive based on some global settings
        core_warnings.high(f'''
{caller()}: no paths were matched against {pp}. This might result in missing data. Likely, the directory you passed is empty.
'''.strip())
        # traceback is useful to figure out what config caused it?
        import traceback

        traceback.print_stack()

    if guess_compression:
        from .kompress import CPath, is_compressed, ZipPath

        # NOTE: wrap is just for backwards compat with vendorized kompress
        # with kompress library, only is_compressed check and Cpath should be enough
        def wrap(p: Path) -> Path:
            if isinstance(p, ZipPath):
                return p
            if p.suffix == '.zip':
                return ZipPath(p)  # type: ignore[return-value]
            if is_compressed(p):
                return CPath(p)
            return p

        paths = [wrap(p) for p in paths]
    return tuple(paths)


@functools.lru_cache(1)
def _magic():
    import magic # type: ignore
    return magic.Magic(mime=True)


# TODO could reuse in pdf module?
import mimetypes # todo do I need init()?
# todo wtf? fastermime thinks it's mime is application/json even if the extension is xz??
# whereas magic detects correctly: application/x-zstd and application/x-xz
def fastermime(path: PathIsh) -> str:
    paths = str(path)
    # mimetypes is faster
    (mime, _) = mimetypes.guess_type(paths)
    if mime is not None:
        return mime
    # magic is slower but returns more stuff
    # TODO Result type?; it's kinda racey, but perhaps better to let the caller decide?
    return _magic().from_file(paths)


Json = Dict[str, Any]


from typing import TypeVar, Callable, Generic

_C = TypeVar('_C')
_R = TypeVar('_R')

# https://stackoverflow.com/a/5192374/706389
class classproperty(Generic[_R]):
    def __init__(self, f: Callable[[_C], _R]) -> None:
        self.f = f

    def __get__(self, obj: None, cls: _C) -> _R:
        return self.f(cls)


# hmm, this doesn't really work with mypy well..
# https://github.com/python/mypy/issues/6244
# class staticproperty(Generic[_R]):
#     def __init__(self, f: Callable[[], _R]) -> None:
#         self.f = f
#
#     def __get__(self) -> _R:
#         return self.f()

# for now just serves documentation purposes... but one day might make it statically verifiable where possible?
# TODO e.g. maybe use opaque mypy alias?
datetime_naive = datetime
datetime_aware = datetime


# TODO deprecate
tzdatetime = datetime_aware


# TODO deprecate (although could be used in modules)
from .compat import fromisoformat as isoparse


import re
# https://stackoverflow.com/a/295466/706389
def get_valid_filename(s: str) -> str:
    s = str(s).strip().replace(' ', '_')
    return re.sub(r'(?u)[^-\w.]', '', s)


from typing import Generic, Sized, Callable


# X = TypeVar('X')
def _warn_iterator(it, f: Any=None):
    emitted = False
    for i in it:
        yield i
        emitted = True
    if not emitted:
        warnings.warn(f"Function {f} didn't emit any data, make sure your config paths are correct")


# TODO ugh, so I want to express something like:
# X = TypeVar('X')
# C = TypeVar('C', bound=Iterable[X])
# _warn_iterable(it: C) -> C
# but apparently I can't??? ugh.
# https://github.com/python/typing/issues/548
# I guess for now overloads are fine...

from typing import overload
X = TypeVar('X')
@overload
def _warn_iterable(it: List[X]    , f: Any=None) -> List[X]    : ...
@overload
def _warn_iterable(it: Iterable[X], f: Any=None) -> Iterable[X]: ...
def _warn_iterable(it, f=None):
    if isinstance(it, Sized):
        sz = len(it)
        if sz == 0:
            warnings.warn(f"Function {f} returned empty container, make sure your config paths are correct")
        return it
    else:
        return _warn_iterator(it, f=f)


# ok, this seems to work...
# https://github.com/python/mypy/issues/1927#issue-167100413
FL = TypeVar('FL', bound=Callable[..., List])
FI = TypeVar('FI', bound=Callable[..., Iterable])

@overload
def warn_if_empty(f: FL) -> FL: ...
@overload
def warn_if_empty(f: FI) -> FI: ...


def warn_if_empty(f):
    from functools import wraps

    @wraps(f)
    def wrapped(*args, **kwargs):
        res = f(*args, **kwargs)
        return _warn_iterable(res, f=f)
    return wrapped


# global state that turns on/off quick stats
# can use the 'quick_stats' contextmanager
# to enable/disable this in cli so that module 'stats'
# functions don't have to implement custom 'quick' logic
QUICK_STATS = False


# in case user wants to use the stats functions/quick option
# elsewhere -- can use this decorator instead of editing
# the global state directly
@contextmanager
def quick_stats():
    global QUICK_STATS
    prev = QUICK_STATS
    try:
        QUICK_STATS = True
        yield
    finally:
        QUICK_STATS = prev


C = TypeVar('C')
Stats = Dict[str, Any]
StatsFun = Callable[[], Stats]
# todo not sure about return type...
def stat(
    func: Union[Callable[[], Iterable[C]], Iterable[C]],
    *,
    quick: bool = False,
    name: Optional[str] = None,
) -> Stats:
    if callable(func):
        fr = func()
        if hasattr(fr, '__enter__') and hasattr(fr, '__exit__'):
            # context managers has Iterable type, but they aren't data providers
            # sadly doesn't look like there is a way to tell from typing annotations
            return {}
        fname = func.__name__
    else:
        # meh. means it's just a list.. not sure how to generate a name then
        fr = func
        fname = f'unnamed_{id(fr)}'
    type_name = type(fr).__name__
    if type_name == 'DataFrame':
        # dynamic, because pandas is an optional dependency..
        df = cast(Any, fr)  # todo ugh, not sure how to annotate properly
        res = dict(
            dtypes=df.dtypes.to_dict(),
            rows=len(df),
        )
    else:
        res = _stat_iterable(fr, quick=quick)

    stat_name = name if name is not None else fname
    return {
        stat_name: res,
    }


def _stat_iterable(it: Iterable[C], quick: bool = False) -> Any:
    from more_itertools import ilen, take, first

    # todo not sure if there is something in more_itertools to compute this?
    total = 0
    errors = 0
    first_item = None
    last_item = None

    def funcit():
        nonlocal errors, first_item, last_item, total
        for x in it:
            total += 1
            if isinstance(x, Exception):
                errors += 1
            else:
                last_item = x
                if first_item is None:
                    first_item = x
            yield x

    eit = funcit()
    count: Any
    if quick or QUICK_STATS:
        initial = take(100, eit)
        count = len(initial)
        if first(eit, None) is not None: # todo can actually be none...
            # haven't exhausted
            count = f'{count}+'
    else:
        count = ilen(eit)

    res = {
        'count': count,
    }

    if total == 0:
        # not sure but I guess a good balance? wouldn't want to throw early here?
        res['warning'] = 'THE ITERABLE RETURNED NO DATA'

    if errors > 0:
        res['errors'] = errors

    def stat_item(item):
        if item is None:
            return None
        if isinstance(item, Path):
            return str(item)
        return guess_datetime(item)

    if (stat_first := stat_item(first_item)) is not None:
        res['first'] = stat_first

    if (stat_last := stat_item(last_item)) is not None:
        res['last'] = stat_last

    return res


def test_stat_iterable() -> None:
    from datetime import datetime, timedelta
    from typing import NamedTuple

    dd = datetime.utcfromtimestamp(123)
    day = timedelta(days=3)

    X = NamedTuple('X', [('x', int), ('d', datetime)])

    def it():
        yield RuntimeError('oops!')
        for i in range(2):
            yield X(x=i, d=dd + day * i)
        yield RuntimeError('bad!')
        for i in range(3):
            yield X(x=i * 10, d=dd + day * (i * 10))
        yield X(x=123, d=dd + day * 50)

    res = _stat_iterable(it())
    assert res['count']  == 1 + 2 + 1 + 3 + 1
    assert res['errors'] == 1 + 1
    assert res['last'] == dd + day * 50


# experimental, not sure about it..
def guess_datetime(x: Any) -> Optional[datetime]:
    # todo hmm implement withoutexception..
    try:
        d = asdict(x)
    except: # noqa: E722 bare except
        return None
    for k, v in d.items():
        if isinstance(v, datetime):
            return v
    return None

def test_guess_datetime() -> None:
    from datetime import datetime
    from dataclasses import dataclass
    from typing import NamedTuple

    dd = isoparse('2021-02-01T12:34:56Z')

    # ugh.. https://github.com/python/mypy/issues/7281
    A = NamedTuple('A', [('x', int)])
    B = NamedTuple('B', [('x', int), ('created', datetime)])

    assert guess_datetime(A(x=4)) is None
    assert guess_datetime(B(x=4, created=dd)) == dd

    @dataclass
    class C:
        a: datetime
        x: int
    assert guess_datetime(C(a=dd, x=435)) == dd
    # TODO not sure what to return when multiple datetime fields?
    # TODO test @property?


def is_namedtuple(thing: Any) -> bool:
    # basic check to see if this is namedtuple-like
    _asdict = getattr(thing, '_asdict', None)
    return (_asdict is not None) and callable(_asdict)


def asdict(thing: Any) -> Json:
    # todo primitive?
    # todo exception?
    if isinstance(thing, dict):
        return thing
    import dataclasses as D
    if D.is_dataclass(thing):
        return D.asdict(thing)
    if is_namedtuple(thing):
        return thing._asdict()
    raise TypeError(f'Could not convert object {thing} to dict')


def assert_subpackage(name: str) -> None:
    # can lead to some unexpected issues if you 'import cachew' which being in my/core directory.. so let's protect against it
    # NOTE: if we use overlay, name can be smth like my.origg.my.core.cachew ...
    assert name == '__main__' or 'my.core' in name, f'Expected module __name__ ({name}) to be __main__ or start with my.core'


from .compat import ParamSpec
_P = ParamSpec('_P')
_T = TypeVar('_T')

# https://stackoverflow.com/a/10436851/706389
from concurrent.futures import Future, Executor
class DummyExecutor(Executor):
    def __init__(self, max_workers: Optional[int]=1) -> None:
        self._shutdown = False
        self._max_workers = max_workers

    if TYPE_CHECKING:
        if sys.version_info[:2] <= (3, 8):
            # 3.8 doesn't support ParamSpec as Callable arg :(
            # and any attempt to type results in incompatible supertype.. so whatever
            def submit(self, fn, *args, **kwargs): ...
        else:
            def submit(self, fn: Callable[_P, _T], /, *args: _P.args, **kwargs: _P.kwargs) -> Future[_T]: ...
    else:
        def submit(self, fn, *args, **kwargs):
            if self._shutdown:
                raise RuntimeError('cannot schedule new futures after shutdown')

            f: Future[Any] = Future()
            try:
                result = fn(*args, **kwargs)
            except KeyboardInterrupt:
                raise
            except BaseException as e:
                f.set_exception(e)
            else:
                f.set_result(result)

            return f

    def shutdown(self, wait: bool=True, **kwargs) -> None:
        self._shutdown = True


# see https://hakibenita.com/python-mypy-exhaustive-checking#exhaustiveness-checking
def assert_never(value: NoReturn) -> NoReturn:
    assert False, f'Unhandled value: {value} ({type(value).__name__})'


def _check_all_hashable(fun):
    # TODO ok, take callable?
    hints = get_type_hints(fun)
    # TODO needs to be defensive like in cachew?
    return_type = hints.get('return')
    # TODO check if None
    origin = get_origin(return_type)  # Iterator etc?
    (arg,) = get_args(return_type)
    # options we wanna handle are simple type on the top level or union
    arg_origin = get_origin(arg)

    if sys.version_info[:2] >= (3, 10):
        is_uniontype = arg_origin is types.UnionType
    else:
        is_uniontype = False

    is_union = arg_origin is Union or is_uniontype
    if is_union:
        to_check = get_args(arg)
    else:
        to_check = (arg,)

    no_hash = [
        t
        for t in to_check
        # seems that objects that have not overridden hash have the attribute but it's set to None
        if getattr(t, '__hash__', None) is None
    ]
    assert len(no_hash) == 0, f'Types {no_hash} are not hashable, this will result in significant performance downgrade for unique_everseen'


_UET = TypeVar('_UET')
_UEU = TypeVar('_UEU')


def unique_everseen(
    fun: Callable[[], Iterable[_UET]],
    key: Optional[Callable[[_UET], _UEU]] = None,
) -> Iterator[_UET]:
    # TODO support normal iterable as well?
    import more_itertools

    # NOTE: it has to take original callable, because otherwise we don't have access to generator type annotations
    iterable = fun()

    if key is None:
        # todo check key return type as well? but it's more likely to be hashable
        if os.environ.get('HPI_CHECK_UNIQUE_EVERSEEN') is not None:
            # TODO return better error here, e.g. if there is no return type it crashes
            _check_all_hashable(fun)

    return more_itertools.unique_everseen(iterable=iterable, key=key)


## legacy imports, keeping them here for backwards compatibility
from functools import cached_property as cproperty
from typing import Literal
from .cachew import mcachew
##
